import numpy as np
from gptcache.utils import import_sbert
from gptcache.embedding.base import BaseEmbedding

import_sbert()

from sentence_transformers import SentenceTransformer  # pylint: disable=C0413


class SBERT(BaseEmbedding):
    """Generate sentence embedding for given text using pretrained models of Sentence Transformers.

    :param model: model name, defaults to 'all-MiniLM-L6-v2'.
    :type model: str

    Example:
        .. code-block:: python

            from gptcache.embedding import SBERT

            test_sentence = 'Hello, world.'
            encoder = SBERT('all-MiniLM-L6-v2')
            embed = encoder.to_embeddings(test_sentence)
    """

    def __init__(self, model: str = "all-MiniLM-L6-v2"):
        self.model = SentenceTransformer(model)
        self.model.eval()
        self.__dimension = None

    def to_embeddings(self, data, **_):
        """Generate embedding given text input

        :param data: text in string.
        :type data: str

        :return: a text embedding in shape of (dim,).
        """
        if not isinstance(data, list):
            data = [data]
        emb = self.model.encode(data).squeeze(0)

        if not self.__dimension:
            self.__dimension = len(emb)
        return np.array(emb).astype("float32")

    @property
    def dimension(self):
        """Embedding dimension.

        :return: embedding dimension
        """
        if not self.__dimension:
            self.__dimension = len(self.to_embeddings("foo"))
        return self.__dimension



from gptcache.utils.lazy_import import LazyImport

summarization = LazyImport(
    "summarization_context",
    globals(),
    "gptcache.processor.context.summarization_context",
)
selective = LazyImport(
    "selective_context", globals(), "gptcache.processor.context.selective_context"
)
concat = LazyImport(
    "concat_context", globals(), "gptcache.processor.context.concat_context"
)


__all__ = [
    "SummarizationContextProcess",
    "SelectiveContextProcess",
    "ConcatContextProcess",
]


def SummarizationContextProcess(model_name=None, tokenizer=None, target_length=512):
    return summarization.SummarizationContextProcess(
        model_name, tokenizer, target_length
    )


def SelectiveContextProcess(
    model_type: str = "gpt2",
    lang: str = "en",
    reduce_ratio: float = 0.35,
    reduce_level: str = "phrase",
):
    return selective.SelectiveContextProcess(
        model_type=model_type,
        lang=lang,
        reduce_ratio=reduce_ratio,
        reduce_level=reduce_level,
    )


def ConcatContextProcess():
    return concat.ConcatContextProcess()



import numpy as np

from gptcache.embedding.base import BaseEmbedding
from gptcache.utils import import_huggingface

import_huggingface()

from transformers import AutoTokenizer, RwkvModel  # pylint: disable=C0413


class Rwkv(BaseEmbedding):
    """Generate sentence embedding for given text using RWKV models.

    :param model: model name, defaults to 'sgugger/rwkv-430M-pile'. Check
      https://huggingface.co/docs/transformers/model_doc/rwkv for more avaliable models.
    :type model: str

    Example:
        .. code-block:: python

            from gptcache.embedding import Rwkv

            test_sentence = 'Hello, world.'
            encoder = Rwkv(model='sgugger/rwkv-430M-pile')
            embed = encoder.to_embeddings(test_sentence)
    """

    def __init__(self, model: str = "sgugger/rwkv-430M-pile"):
        self.model = RwkvModel.from_pretrained(model)
        self.model.eval()

        self.tokenizer = AutoTokenizer.from_pretrained(model)
        try:
            self.__dimension = self.model.config.hidden_size
        except Exception:  # pylint: disable=W0703
            from transformers import AutoConfig  # pylint: disable=C0415

            config = AutoConfig.from_pretrained(model)
            self.__dimension = config.hidden_size

    def to_embeddings(self, data, **_):
        """Generate embedding given text input

        :param data: text in string.
        :type data: str

        :return: a text embedding in shape of (dim,).
        """
        inputs = self.tokenizer(data, return_tensors="pt")
        outputs = self.model(inputs["input_ids"])
        emb = outputs.last_hidden_state[0, 0, :].detach().numpy()
        return np.array(emb).astype("float32")

    @property
    def dimension(self):
        """Embedding dimension.

        :return: embedding dimension
        """
        return self.__dimension



# pylint: disable=wrong-import-position
from abc import ABC, abstractmethod
from typing import List

from gptcache.manager.eviction.base import EvictionBase


class DistributedEviction(EvictionBase, ABC):
    """
    Base class for Distributed Eviction Strategy.
    """

    @abstractmethod
    def put(self, objs: List[str]):
        pass

    @abstractmethod
    def get(self, obj: str):
        pass

    @property
    @abstractmethod
    def policy(self) -> str:
        pass


class NoOpEviction(EvictionBase):
    """eviction: No Op Eviction Strategy. This is used when Eviction is managed internally
    by the Databases such as Redis or memcached and no eviction is required to perform.

    """

    @property
    def policy(self) -> str:
        return ""

    def __init__(self, **kwargs):
        pass

    def put(self, objs: List[str]):
        pass

    def get(self, obj: str):
        pass



from gptcache.utils import import_tiktoken

_encoding = None


def _get_encoding():
    global _encoding
    if _encoding is None:
        import_tiktoken()
        import tiktoken  # pylint: disable=C0415
        _encoding = tiktoken.get_encoding("cl100k_base")
    return _encoding


def token_counter(text):
    """Token Counter"""
    num_tokens = len(_get_encoding().encode(text))
    return num_tokens



import logging

import gptcache

FORMAT = '%(asctime)s - %(thread)d - %(filename)s-%(module)s:%(lineno)s - %(levelname)s: %(message)s'
logging.basicConfig(format=FORMAT)

gptcache_log = logging.getLogger(f'gptcache:{gptcache.__version__}')



from typing import Union, Any

from gptcache.embedding.base import BaseEmbedding
from gptcache.utils import import_uform, import_pillow
from gptcache.utils.error import ParamError

import_pillow()
import_uform()

from uform import TritonClient, get_model  # pylint: disable=C0413 # nopep8
from PIL import Image  # pylint: disable=C0413 # nopep8


class UForm(BaseEmbedding):
    """Generate multi-modal embeddings using pretrained models from UForm.

    :param model: model name, defaults to 'unum-cloud/uform-vl-english'.
    :type model: str
    :param embedding_type: type of embedding, defaults to 'text'. options: text, image
    :type embedding_type: str

    Example:
        .. code-block:: python

            from gptcache.embedding import UForm

            test_sentence = 'Hello, world.'
            encoder = UForm(model='unum-cloud/uform-vl-english')
            embed = encoder.to_embeddings(test_sentence)

            test_sentence = '什么是Github'
            encoder = UForm(model='unum-cloud/uform-vl-multilingual')
            embed = encoder.to_embeddings(test_sentence)
    """

    def __init__(self, model: Union[str, TritonClient] = "unum-cloud/uform-vl-english", embedding_type: str = "text"):
        if isinstance(model, str):
            self.__model = get_model(model)
        else:
            self.__model = model
        self.__embedding_type = embedding_type
        if embedding_type == "text":
            self.__dimension = self.__model.text_encoder.proj.out_features
        elif embedding_type == "image":
            self.__dimension = self.__model.img_encoder.proj.out_features
        else:
            raise ParamError(f"Unknown embedding type: {embedding_type}")

    def to_embeddings(self, data: Any, **_):
        """Generate embedding given text input or a path to a file.

        :param data: text in string, or a path to an image file.
        :type data: str

        :return: an embedding in shape of (dim,).
        """
        if self.__embedding_type == "image":
            data = Image.open(data)
            data = self.__model.preprocess_image(data)
            emb = self.__model.encode_image(data)
        else:
            data = self.__model.preprocess_text(data)
            emb = self.__model.encode_text(data)
        return emb.detach().numpy().flatten()

    @property
    def dimension(self):
        """Embedding dimension.

        :return: embedding dimension
        """
        return self.__dimension



import os
from pathlib import Path
from typing import Union, Callable

from gptcache.manager import CacheBase, VectorBase, ObjectBase
from gptcache.manager.data_manager import SSDataManager, MapDataManager
from gptcache.manager.eviction import EvictionBase
from gptcache.utils.log import gptcache_log


def manager_factory(manager="map",
                    data_dir="./",
                    max_size=1000,
                    eviction_manager: str = "memory",
                    get_data_container: Callable = None,
                    scalar_params=None,
                    vector_params=None,
                    object_params=None,
                    eviction_params=None
                    ):
    """Factory of DataManager.
       By using this factory method, you only need to specify the root directory of the data,
       and it can automatically manage all the local files.

    :param manager: Type of DataManager. Supports: Map, or {scalar_name},{vector_name}
                    or {scalar_name},{vector_name},{object_name}
    :type manager: str
    :param data_dir: Root path for data storage.
    :type data_dir: str
    :param max_size: the max size for the LRU cache in MapDataManager, defaults to 1000.
    :type max_size: int
    :param eviction_manager: The eviction manager, defaults to "memory".
                             It supports "memory" and "redis" and 'no_op_eviction'.
    :type eviction_manager:  str
    :param get_data_container: a Callable to get the data container, defaults to None.
    :type get_data_container:  Callable

    :param scalar_params: Params of scalar storage.
    :type scalar_params:  dict

    :param vector_params: Params of vector storage.
    :type vector_params:  dict

    :param object_params: Params of object storage.
    :type object_params:  dict

    :param eviction_params: Params of eviction.
    :type eviction_params:  dict
    :return: SSDataManager or MapDataManager.

    Example:
        .. code-block:: python

            from gptcache.manager import manager_factory

            data_manager = manager_factory("sqlite,faiss", data_dir="./workspace", vector_params={"dimension": 128})

            # or using manager factory enabled with redis cache instead of in-memory cache
            from gptcache.manager import manager_factory
            data_manager = manager_factory("redis,faiss",
                                           eviction_manager="redis",
                                           scalar_params={"maxmemory": "2mb",
                                                          "policy": "allkeys-lru"
                                                          },
                                           vector_params={"dimension": 128},
                                           eviction_params=dict(url="redis://localhost:6379")
                                           )
    """

    Path(data_dir).mkdir(parents=True, exist_ok=True)
    manager = manager.lower()

    if manager == "map":
        return MapDataManager(os.path.join(data_dir, "data_map.txt"), max_size, get_data_container)

    db_infos = manager.split(",")
    if len(db_infos) not in [2, 3]:
        raise RuntimeError(
            "Error manager format: %s, the correct is \"{scalar},{vector},{object}\", object is optional" % manager)

    if len(db_infos) == 2:
        db_infos.append("")
    scalar, vector, obj = db_infos

    if scalar_params is None:
        scalar_params = {}
    if scalar == "sqlite":
        scalar_params["sql_url"] = "sqlite:///" + os.path.join(data_dir, "sqlite.db")
    s = CacheBase(name=scalar, **scalar_params)

    if vector_params is None:
        vector_params = {}
    local_vector_type = ["faiss", "hnswlib", "docarray"]
    if vector in local_vector_type:
        vector_params["index_path"] = os.path.join(data_dir, f"{vector}.index")
    elif vector == "milvus" and vector_params.get("local_mode", False) is True:
        vector_params["local_data"] = os.path.join(data_dir, "milvus_data")
    v = VectorBase(name=vector, **vector_params)

    if object_params is None:
        object_params = {}
    if obj == "local":
        object_params["path"] = os.path.join(data_dir, "local_obj")
    o = ObjectBase(name=obj, **object_params) if obj else None

    if eviction_params is None:
        eviction_params = {}

    if  scalar == "redis" and eviction_manager == "redis":
        # if cache manager and eviction manager are both redis, we use no op redis to avoid redundant operations
        eviction_manager = "no_op_eviction"
        gptcache_log.info("Since Scalar Storage and Eviction manager are both redis, "
                          "no_op_eviction will be used to avoid redundant operations.")
        s.init_eviction_params(
            maxmemory=eviction_params.get("maxmemory", scalar_params.get("maxmemory")),
            policy=eviction_params.get("policy", scalar_params.get("policy")),
            ttl=eviction_params.get("ttl", scalar_params.get("ttl")),
            maxmemory_samples=eviction_params.get("maxmemory_samples", scalar_params.get("maxmemory_samples")),
        )

    e = EvictionBase(
        name=eviction_manager,
        **eviction_params
    )
    return get_data_manager(s, v, o, e)


def get_data_manager(
        cache_base: Union[CacheBase, str] = None,
        vector_base: Union[VectorBase, str] = None,
        object_base: Union[ObjectBase, str] = None,
        eviction_base: Union[EvictionBase, str] = None,
        max_size: int = 1000,
        clean_size=None,
        eviction: str = "LRU",
        data_path: str = "data_map.txt",
        get_data_container: Callable = None,
):
    """Generate `SSDataManager` (with `cache_base`, `vector_base`, `max_size`, `clean_size` and `eviction` params),
       or `MAPDataManager` (with `data_path`, `max_size` and `get_data_container` params) to manager the data.

    :param cache_base: a CacheBase object, or the name of the cache storage, it is support 'sqlite', 'duckdb', 'postgresql',
                       'mysql', 'mariadb', 'sqlserver' and  'oracle' now.
    :type cache_base: :class:`CacheBase` or str
    :param vector_base: a VectorBase object, or the name of the vector storage, it is support 'milvus', 'faiss' and
                        'chromadb' now.
    :type vector_base: :class:`VectorBase` or str
    :param object_base: a object storage, supports local path and s3.
    :type object_base: :class:`ObjectBase` or str
    :param max_size: the max size for the LRU cache in MapDataManager, defaults to 1000.
    :type max_size: int
    :param eviction_base: a EvictionBase object, or the name of the eviction, it supports:
     - 'memory'
     - 'redis'
     - 'no_op_eviction'.
    :type eviction_base: :class:`EvictionBase` or str
    :param clean_size: the clean size for the LRU cache in MapDataManager, defaults to None.
    :type clean_size: int
    :param eviction: the eviction policy for the LRU cache in MapDataManager, defaults to 'LRU'.
    :type eviction: str
    :param data_path: the path to save the map data, defaults to 'data_map.txt'.
    :type data_path:  str
    :param get_data_container: a Callable to get the data container, defaults to None.
    :type get_data_container:  Callable


    :return: SSDataManager or MapDataManager.

    Example:
        .. code-block:: python

            from gptcache.manager import get_data_manager, CacheBase, VectorBase

            data_manager = get_data_manager(CacheBase('sqlite'), VectorBase('faiss', dimension=128))

            # or using manager factory enabled with redis cache instead of in-memory cache
            # example 1: using redis eviction base with sqlite cache base
            data_manager = get_data_manager(cache_base=CacheBase("sqlite"),
                                        vector_base=VectorBase("faiss", dimension=onnx.dimension),
                                        eviction_base=EvictionBase("redis",  maxmemory="0", policy="noeviction", ttl=1))

           # example 2: using redis eviction base with redis cache base
           # here no_op_eviction is used since `redis` cache base already handles evictions internally
            data_manager = get_data_manager(cache_base=CacheBase("redis",  maxmemory="0", policy="noeviction", ttl=1),
                                        vector_base=VectorBase("faiss", dimension=onnx.dimension),
                                        eviction_base=EvictionBase("no_op_eviction"))
    """
    if not cache_base and not vector_base:
        return MapDataManager(data_path, max_size, get_data_container)

    if isinstance(cache_base, str):
        cache_base = CacheBase(name=cache_base)
    if isinstance(vector_base, str):
        vector_base = VectorBase(name=vector_base)
    if isinstance(object_base, str):
        object_base = ObjectBase(name=object_base)
    if isinstance(eviction_base, str):
        eviction_base = EvictionBase(name=eviction_base)
    assert cache_base and vector_base
    return SSDataManager(cache_base, vector_base, object_base, eviction_base, max_size, clean_size, eviction)



from typing import List
from uuid import uuid4
import numpy as np

from gptcache.utils import import_pymilvus
from gptcache.utils.log import gptcache_log
from gptcache.manager.vector_data.base import VectorBase, VectorData


import_pymilvus()

from pymilvus import (  # pylint: disable=C0413
    connections,
    utility,
    FieldSchema,
    DataType,
    CollectionSchema,
    Collection,
    MilvusException,
)


class Milvus(VectorBase):
    """vector store: Milvus

    :param host: the host for Milvus vector database, defaults to 'localhost'.
    :type host: str
    :param port: the port for Milvus vector database, defaults to '19530'.
    :type port: str
    :param user: the user for Zilliz Cloud, defaults to "".
    :type user: str
    :param password: the password for Zilliz Cloud, defaults to "".
    :type password: str
    :param secure: whether it is https with Zilliz Cloud, defaults to False.
    :type secures: bool
    :param collection_name: the name of the collection for Milvus vector database, defaults to 'gptcache'.
    :type collection_name: str
    :param dimension: the dimension of the vector, defaults to 0.
    :type dimension: int
    :param top_k: the number of the vectors results to return, defaults to 1.
    :type top_k: int
    :param index_params: the index parameters for Milvus, defaults to the HNSW index: {'metric_type': 'L2', 'index_type': 'HNSW', 'params': {'M':
                         8, 'efConstruction': 64}}.
    :type index_params: dict
    :param local_mode: if true, will start a local milvus server.
    :type local_mode: bool
    :param local_data: required when local_mode is True.
    :type local_data: str
    """

    SEARCH_PARAM = {
        "IVF_FLAT": {"metric_type": "L2", "params": {"nprobe": 10}},
        "IVF_SQ8": {"metric_type": "L2", "params": {"nprobe": 10}},
        "IVF_PQ": {"metric_type": "L2", "params": {"nprobe": 10}},
        "HNSW": {"metric_type": "L2", "params": {"ef": 10}},
        "RHNSW_FLAT": {"metric_type": "L2", "params": {"ef": 10}},
        "RHNSW_SQ": {"metric_type": "L2", "params": {"ef": 10}},
        "RHNSW_PQ": {"metric_type": "L2", "params": {"ef": 10}},
        "IVF_HNSW": {"metric_type": "L2", "params": {"nprobe": 10, "ef": 10}},
        "ANNOY": {"metric_type": "L2", "params": {"search_k": 10}},
        "AUTOINDEX": {"metric_type": "L2", "params": {}},
    }

    def __init__(
        self,
        host: str = "localhost",
        port: str = "19530",
        user: str = "",
        password: str = "",
        secure: bool = False,
        collection_name: str = "gptcache",
        dimension: int = 0,
        top_k: int = 1,
        index_params: dict = None,
        search_params: dict = None,
        local_mode: bool = False,
        local_data: str = "./milvus_data"
    ):
        if dimension <= 0:
            raise ValueError(
                f"invalid `dim` param: {dimension} in the Milvus vector store."
            )
        self._local_mode = local_mode
        self._local_data = local_data
        self.dimension = dimension
        self.top_k = top_k
        self.index_params = index_params
        if self._local_mode:
            self._create_local(port, local_data)
        self._connect(host, port, user, password, secure)
        self._create_collection(collection_name)
        self.search_params = (
            search_params or self.SEARCH_PARAM[self.index_params["index_type"]]
        )

    def _create_local(self, port, local_data):
        from gptcache.utils import import_milvus_lite  # pylint: disable=import-outside-toplevel
        import_milvus_lite()
        from milvus import MilvusServer  # pylint: disable=import-outside-toplevel
        self._server = MilvusServer()
        self._server.set_base_dir(local_data)
        self._server.listen_port = int(port)
        self._server.start()

    def _connect(self, host, port, user, password, secure):
        try:
            i = [
                connections.get_connection_addr(x[0])
                for x in connections.list_connections()
            ].index({"host": host, "port": port})
            self.alias = connections.list_connections()[i][0]
        except ValueError:
            # Connect to the Milvus instance using the passed in Environment variables
            self.alias = uuid4().hex
            connections.connect(
                alias=self.alias,
                host=host,
                port=port,
                user=user,  # type: ignore
                password=password,  # type: ignore
                secure=secure,
                timeout=10
            )

    def _create_collection(self, collection_name):
        if not utility.has_collection(collection_name, using=self.alias):
            schema = [
                FieldSchema(
                    name="id",
                    dtype=DataType.INT64,
                    is_primary=True,
                    auto_id=False,
                ),
                FieldSchema(
                    name="embedding", dtype=DataType.FLOAT_VECTOR, dim=self.dimension
                ),
            ]
            schema = CollectionSchema(schema)
            self.col = Collection(
                collection_name,
                schema=schema,
                consistency_level="Strong",
                using=self.alias,
            )
        else:
            gptcache_log.warning("The %s collection already exists, and it will be used directly.", collection_name)
            self.col = Collection(
                collection_name, consistency_level="Strong", using=self.alias
            )

        if len(self.col.indexes) == 0:
            try:
                gptcache_log.info("Attempting creation of Milvus index.")
                self.col.create_index("embedding", index_params=self.index_params)
                gptcache_log.info("Creation of Milvus index successful.")
            except MilvusException as e:
                gptcache_log.warning("Error with building index: %s, and attempting creation of default index.", e)
                i_p = {"metric_type": "L2", "index_type": "AUTOINDEX", "params": {}}
                self.col.create_index("embedding", index_params=i_p)
                self.index_params = i_p
        else:
            self.index_params = self.col.indexes[0].to_dict()["index_param"]

        self.col.load()

    def mul_add(self, datas: List[VectorData]):
        data_array, id_array = map(list, zip(*((data.data, data.id) for data in datas)))
        np_data = np.array(data_array).astype("float32")
        entities = [id_array, np_data]
        self.col.insert(entities)

    def search(self, data: np.ndarray, top_k: int = -1):
        if top_k == -1:
            top_k = self.top_k
        search_result = self.col.search(
            data=data.reshape(1, -1).tolist(),
            anns_field="embedding",
            param=self.search_params,
            limit=top_k,
        )
        return list(zip(search_result[0].distances, search_result[0].ids))

    def delete(self, ids):
        del_ids = ",".join([str(x) for x in ids])
        self.col.delete(f"id in [{del_ids}]")

    def rebuild(self, ids=None):  # pylint: disable=unused-argument
        self.col.compact()

    def flush(self):
        self.col.flush(_async=True)

    def close(self):
        self.flush()
        if self._local_mode:
            self._server.stop()

    def get_embeddings(self, data_id: int):
        vec_emb = self.col.query(
            expr=f"id == {data_id}",
            output_fields=["embedding"],
        )
        if len(vec_emb) < 1:
            return None
        vec_emb = np.asarray(vec_emb[0]["embedding"], dtype="float32")
        return vec_emb

    def update_embeddings(self, data_id: int, emb: np.ndarray):
        self.col.delete(f"id in [{data_id}]")
        data = [
            [data_id],
            np.expand_dims(emb, axis=0),
        ]
        self.col.insert(data)



from gptcache.utils.error import NotFoundError


class ObjectBase:
    """
    ObjectBase to manager the object storage.

    Generate specific ObjectStorage with the configuration. For example, setting for
       `ObjectBase` (with `name`) to manage LocalObjectStorage, S3 object storage.

    :param name: the name of the object storage, it is support 'local', 's3'.
    :type name: str
    :param path: the cache root of the LocalObjectStorage.
    :type path: str

    :param bucket: the bucket of s3.
    :type bucket: str
    :param path_prefix: s3 object prefix.
    :type path_prefix: str
    :param access_key: the access_key of s3.
    :type access_key: str
    :param secret_key: the secret_key of s3.
    :type secret_key: str

    :return: ObjectStorage.

    Example:
        .. code-block:: python

            from gptcache.manager import ObjectBase

            obj_storage = ObjectBase('local', path='./')
    """

    def __init__(self):
        raise EnvironmentError(
            "CacheBase is designed to be instantiated, please using the `CacheBase.get(name)`."
        )

    @staticmethod
    def get(name, **kwargs):
        if name == "local":
            from gptcache.manager.object_data.local_storage import LocalObjectStorage  # pylint: disable=import-outside-toplevel
            object_base = LocalObjectStorage(kwargs.get("path", "./local_obj"))
        elif name == "s3":
            from gptcache.manager.object_data.s3_storage import S3Storage  # pylint: disable=import-outside-toplevel
            object_base = S3Storage(kwargs.get("path_prefix"), kwargs.get("bucket"),
                                    kwargs.get("access_key"), kwargs.get("secret_key"),
                                    kwargs.get("endpoint"))
        else:
            raise NotFoundError("object store", name)
        return object_base



from gptcache.utils.error import CacheError
from gptcache.adapter.adapter import adapt
from gptcache.utils import import_replicate
from gptcache.manager.scalar_data.base import DataType, Question, Answer

import_replicate()

import replicate  # pylint: disable=C0413


class Client(replicate.client.Client):
    """replicate.client.Client Wrapper

    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.adapter import replicate
            from gptcache.processor.pre import get_input_str
            from gptcache.embedding import Timm
            from gptcache.similarity_evaluation.onnx import OnnxModelEvaluation

            # init gptcache
            timm = Timm('resnet18')
            cache_base = CacheBase('sqlite')
            vector_base = VectorBase('faiss', dimension=timm.dimension)
            object_base = ObjectBase('local', path='./objects')
            data_manager = get_data_manager(cache_base, vector_base, object_base)

            cache.init(
                pre_embedding_func=get_input_image_file_name,
                data_manager=data_manager,
                embedding_func=timm.to_embeddings,
                similarity_evaluation=OnnxModelEvaluation()
                )

            # run replicate clinet with gptcache
            output = replicate.run(
                        "andreasjansson/blip-2:4b32258c42e9efd4288bb9910bc532a69727f9acd26aa08e175713a0a857a608",
                        input={"image": open("/path/to/merlion.png", "rb"),
                               "question": "Which city is this photo taken on?"}
                    )
    """
    def run(self, model_version: str, **kwargs):
        if "input" in kwargs and "question" in kwargs["input"] and "image" in kwargs["input"]:
            cache_context = {"deps": [
                {"name": "text", "data": kwargs["input"]["question"], "dep_type": DataType.STR},
                {"name": "image", "data": kwargs["input"]["image"].name, "dep_type": DataType.STR},
                ]}
        else:
            cache_context = {}
        def llm_handler(*llm_args, **llm_kwargs):
            try:
                return replicate.run(*llm_args, **llm_kwargs)
            except Exception as e:
                raise CacheError("replicate error") from e

        def cache_data_convert(cache_data):
            return cache_data

        def update_cache_callback(llm_data, update_cache_func, *args, **kwargs):  # pylint: disable=unused-argument
            if "input" in kwargs and "question" in kwargs["input"] and "image" in kwargs["input"]:
                question = Question.from_dict({
                    "content": "pre_embedding_data",
                    "deps": [
                        {"name": "text", "data": kwargs["input"]["question"], "dep_type": DataType.STR},
                        {"name": "image", "data": kwargs["input"]["image"].name, "dep_type": DataType.STR},
                    ]
                })
                update_cache_func(Answer(llm_data, DataType.STR), question=question)
            else:
                update_cache_func(llm_data)
            return llm_data


        return adapt(
            llm_handler, cache_data_convert, update_cache_callback,
            model_version=model_version,
            cache_context=cache_context,
            require_object_store=False,
            **kwargs
        )


default_client = Client()
run = default_client.run



import numpy as np
from typing import Dict, Any

from gptcache.similarity_evaluation.distance import SearchDistanceEvaluation
from gptcache.manager.vector_data.base import VectorBase


def euclidean_distance_calculate(vec_l: np.array, vec_r: np.array):
    return np.sum((vec_l - vec_r)**2)

class KReciprocalEvaluation(SearchDistanceEvaluation):
    """Using K Reciprocal to evaluate sentences pair similarity.

    This evaluator borrows popular reranking method K-reprocical reranking for similarity evaluation. K-reciprocal relation refers to the mutual
    nearest neighbor relationship between two embeddings, where each embedding is the K nearest neighbor of the other based on a given distance
    metric.  This evaluator checks whether the query embedding is in candidate cache embedding's `top_k` nearest neighbors. If query embedding
    is not candidate's `top_k` neighbors, the pair will be considered as dissimilar pair. Otherwise, their distance will be kept and continue
    for a `SearchDistanceEvaluation` check.  `max_distance` is used to bound this distance to make it between [0-`max_distance`]. `positive` is
    used to indicate this distance is directly proportional to the similarity of two entites. If `positive` is set `False`,
    `max_distance` will be used to substract this distance to get the final score.

    :param vectordb: vector database to retrieval embeddings to test k-reciprocal relationship.
    :type vectordb: gptcache.manager.vector_data.base.VectorBase
    :param top_k: for each retievaled candidates, this method need to test if the query is top-k of candidate.
    :type top_k: int
    :param max_distance: the bound of maximum distance.
    :type max_distance: float
    :param positive: if the larger distance indicates more similar of two entities, It is True. Otherwise it is False.
    :type positive: bool

    Example:
        .. code-block:: python

            from gptcache.similarity_evaluation import KReciprocalEvaluation
            from gptcache.manager.vector_data.faiss import Faiss
            from gptcache.manager.vector_data.base import VectorData
            import numpy as np

            faiss = Faiss('./none', 3, 10)
            cached_data = np.array([0.57735027, 0.57735027, 0.57735027])
            faiss.mul_add([VectorData(id=0, data=cached_data)])
            evaluation = KReciprocalEvaluation(vectordb=faiss, top_k=2, max_distance = 4.0, positive=False)
            query = np.array([0.61396013, 0.55814557, 0.55814557])
            score = evaluation.evaluation(
                {
                    'question': 'question1',
                    'embedding': query
                },
                {
                    'question': 'question2',
                    'embedding': cached_data
                }
            )
    """


    def __init__(self, vectordb: VectorBase, top_k: int = 3, max_distance: float = 4.0, positive: bool=False):
        super().__init__(max_distance, positive)
        self.vectordb = vectordb
        self.top_k = top_k

    @staticmethod
    def normalize(vec: np.ndarray):
        """Normalize the input vector.

        :param vec: numpy vector needs to normalize.
        :type vec: numpy.array

        :return: normalized vector.
        """
        magnitude = np.linalg.norm(vec)
        normalized_v = vec / magnitude
        return normalized_v


    def evaluation(
        self, src_dict: Dict[str, Any], cache_dict: Dict[str, Any], **_
    ) -> float:
        """Evaluate the similarity score of pair.

        :param src_dict: the query dictionary to evaluate with cache.
        :type src_dict: Dict
        :param cache_dict: the cache dictionary.
        :type cache_dict: Dict

        :return: evaluation score.
        """
        src_question = src_dict['question']
        cache_question = cache_dict['question']
        if src_question == cache_question:
            return 1
        query_emb = self.normalize(src_dict['embedding'])
        candidates = self.vectordb.search(cache_dict['embedding'], self.top_k + 1)
        euc_dist = euclidean_distance_calculate(query_emb, cache_dict['embedding'])
        if euc_dist > candidates[-1][0]:
            euc_dist = self.range()[1]

        result_dict = {}
        result_dict['search_result'] = (euc_dist, None)
        return super().evaluation(None, result_dict)





__all__ = [
    "import_pymilvus",
    "import_milvus_lite",
    "import_sbert",
    "import_cohere",
    "import_fasttext",
    "import_huggingface",
    "import_uform",
    "import_usearch",
    "import_torch",
    "import_huggingface_hub",
    "import_onnxruntime",
    "import_faiss",
    "import_hnswlib",
    "import_chromadb",
    "import_sqlalchemy",
    "import_sql_client",
    "import_mongodb",
    "import_pydantic",
    "import_langchain",
    "import_pillow",
    "import_boto3",
    "import_diffusers",
    "import_torchaudio",
    "import_torchvision",
    "import_timm",
    "import_vit",
    "import_replicate",
    "import_stability",
    "import_scipy",
    "import_llama_cpp_python",
    "import_ruamel",
    "import_selective_context",
    "import_httpx",
    "import_openai",
    "import_docarray",
    "softmax",
    "import_paddle",
    "import_paddlenlp",
    "import_tiktoken",
    "import_fastapi",
    "import_redis",
    "import_qdrant",
    "import_weaviate",
    ]

import importlib.util
from typing import Optional

from gptcache.utils.dependency_control import prompt_install
from gptcache.utils.softmax import softmax  # pylint: disable=unused-argument


def _check_library(libname: str, prompt: bool = True, package: Optional[str] = None):
    is_avail = False
    if importlib.util.find_spec(libname):
        is_avail = True
    if not is_avail and prompt:
        prompt_install(package if package else libname)
    return is_avail


def import_pymilvus():
    _check_library("pymilvus")


def import_milvus_lite():
    _check_library("milvus")


def import_qdrant():
    _check_library("qdrant_client")


def import_sbert():
    _check_library("sentence_transformers", package="sentence-transformers")


def import_cohere():
    _check_library("cohere")


def import_fasttext():
    _check_library("fasttext", package="fasttext==0.9.2")


def import_huggingface():
    _check_library("transformers")


def import_uform():
    _check_library("uform", package="uform==0.2.1")


def import_usearch():
    _check_library("usearch", package="usearch==0.22.3")


def import_torch():
    _check_library("torch")


def import_huggingface_hub():
    _check_library("huggingface_hub", package="huggingface-hub")


def import_onnxruntime():
    _check_library("onnxruntime", package="onnxruntime==1.14.1")


def import_faiss():
    _check_library("faiss", package="faiss-cpu")


def import_hnswlib():
    _check_library("hnswlib")


def import_chromadb():
    _check_library("chromadb", package="chromadb==0.3.26")


def import_sqlalchemy():
    _check_library("sqlalchemy")


def import_postgresql():
    _check_library("psycopg2", package="psycopg2-binary")


def import_pymysql():
    _check_library("pymysql")


# `brew install unixodbc` in mac
# and install PyODBC driver.
def import_pyodbc():
    _check_library("pyodbc")


# install cx-Oracle driver.
def import_cxoracle():
    _check_library("cx_Oracle")


def import_duckdb():
    _check_library("duckdb", package="duckdb")
    _check_library("duckdb-engine", package="duckdb-engine")


def import_sql_client(db_name):
    if db_name == "postgresql":
        import_postgresql()
    elif db_name in ["mysql", "mariadb"]:
        import_pymysql()
    elif db_name == "sqlserver":
        import_pyodbc()
    elif db_name == "oracle":
        import_cxoracle()
    elif db_name == "duckdb":
        import_duckdb()


def import_mongodb():
    _check_library("pymongo")
    _check_library("mongoengine")


def import_pydantic():
    _check_library("pydantic")


def import_langchain():
    _check_library("langchain")


def import_pillow():
    _check_library("PIL", package="pillow")


def import_boto3():
    _check_library("boto3")


def import_diffusers():
    _check_library("diffusers")


def import_torchaudio():
    _check_library("torchaudio")


def import_torchvision():
    _check_library("torchvision")


def import_timm():
    _check_library("timm", package="timm")


def import_vit():
    _check_library("vit", package="vit")


def import_replicate():
    _check_library("replicate")


def import_stability():
    _check_library("stability_sdk", package="stability-sdk")


def import_scipy():
    _check_library("scipy")


def import_llama_cpp_python():
    _check_library("llama_cpp", package="llama-cpp-python")


def import_ruamel():
    _check_library("ruamel-yaml")


def import_selective_context():
    _check_library("selective_context")


def import_httpx():
    _check_library("httpx")


def import_openai():
    _check_library("openai", package="openai==0.28.1")


def import_docarray():
    _check_library("docarray")


def import_paddle():
    prompt_install("protobuf==3.20.0")
    _check_library("paddlepaddle")


def import_paddlenlp():
    _check_library("paddlenlp")


def import_tiktoken():
    _check_library("tiktoken")


def import_fastapi():
    _check_library("uvicorn", package="'uvicorn[standard]'")
    _check_library("fastapi")


def import_redis():
    _check_library("redis")
    _check_library("redis_om", package="redis-om")


def import_starlette():
    _check_library("starlette")


def import_weaviate():
    _check_library("weaviate-client")



from datetime import datetime
from typing import Tuple, Dict, Any

from gptcache.adapter.api import _get_eval
from gptcache.similarity_evaluation import SimilarityEvaluation


class TimeEvaluation(SimilarityEvaluation):
    """Add time dimension restrictions on the basis of other Evaluation,
    for example, only use the cache within 1 day from the current time,
    and filter out the previous cache.

    :param evaluation: Similarity evaluation, like distance/onnx.
    :param evaluation_config: Similarity evaluation config.
    :param time_range: Time range, time unit: s

    Example:
        .. code-block:: python

            import datetime

            from gptcache.manager.scalar_data.base import CacheData
            from gptcache.similarity_evaluation import TimeEvaluation

            evaluation = TimeEvaluation(evaluation="distance", time_range=86400)

            similarity = eval.evaluation(
                {},
                {
                    "search_result": (3.5, None),
                    "cache_data": CacheData("a", "b", create_on=datetime.datetime.now()),
                },
            )
            # 0.5

    """

    def __init__(self, evaluation: str, evaluation_config=None, time_range: float = 86400.0):
        if evaluation_config is None:
            evaluation_config = {}
        self._eval = _get_eval(evaluation, evaluation_config)
        self._time_range = time_range

    def evaluation(self, src_dict: Dict[str, Any], cache_dict: Dict[str, Any], **kwargs) -> float:
        cache_data = cache_dict.get("cache_data", None)
        if not cache_data or not cache_data.create_on:
            return self.range()[0]
        delta_time = datetime.now().timestamp() - cache_data.create_on.timestamp()
        if delta_time > self._time_range:
            return self.range()[0]
        return self._eval.evaluation(src_dict, cache_dict, **kwargs)

    def range(self) -> Tuple[float, float]:
        return self._eval.range()




# pylint: disable=unused-argument
def check_hit_session(cur_session_id: str, cache_session_ids: list, cache_questions: list, cache_answer: str):
    """
    Check if the sesion result meets the hit requirement.

    :param cur_session_id: the name of the current session.
    :type cur_session_id: str
    :param cache_session_ids: a list of session names for caching the same content if you are using map as a data management method.
                              Otherwise a list of session names for similar content and same answer.
    :type cache_session_ids: list
    :param cache_question: a list with one question which same as the you asked if you use a map as a data management method.
                           Otherwise it is a list that is similar to the question you asked with the same answer,
                           and it is correspondence with cache_session_ids.
    :type cache_question: list
    :param cache_answer: the content of the cached answer.
    :param cache_answer: str

    :return: True or False
    """
    return cur_session_id not in cache_session_ids



from typing import Dict, Tuple, Any

import numpy as np

from gptcache.similarity_evaluation import SimilarityEvaluation


class NumpyNormEvaluation(SimilarityEvaluation):
    """Using Numpy norm to evaluate sentences pair similarity.

    This evaluator calculate the L2 distance of two embeddings for similarity check. if `enable_normal` is True,
    both query embedding and cache embedding will be normalized. Note normalized distance will substracted by
    maximum distance so it will be positively correlated with the similarity.

    :param enable_normal: whether to normalize the embedding, defaults to False.
    :type enable_normal: bool
    :param question_embedding_function: optional, a function to generate question embedding
    :type question_embedding_function: function

    Example:
        .. code-block:: python

            from gptcache.similarity_evaluation import NumpyNormEvaluation
            import numpy as np

            evaluation = NumpyNormEvaluation()
            score = evaluation.evaluation(
                {
                    'question': 'What is color of sky?'
                    'embedding': np.array([-0.5, -0.5])
                },
                {
                    'question': 'What is the color of sky?'
                    'embedding': np.array([-0.49, -0.51])
                }
            )
    """

    def __init__(self, enable_normal: bool = True, question_embedding_function=None):
        self.enable_normal = enable_normal
        self.question_encoder = question_embedding_function

    @staticmethod
    def normalize(vec: np.ndarray):
        """Normalize the input vector.

        :param vec: numpy vector needs to normalize.
        :type vec: numpy.array

        :return: normalized vector.
        """
        magnitude = np.linalg.norm(vec)
        normalized_v = vec / magnitude
        return normalized_v

    def evaluation(
        self, src_dict: Dict[str, Any], cache_dict: Dict[str, Any], **_
    ) -> float:
        """Evaluate the similarity score of pair.

        :param src_dict: the query dictionary to evaluate with cache.
        :type src_dict: Dict
        :param cache_dict: the cache dictionary.
        :type cache_dict: Dict

        :return: evaluation score.
        """
        if 'question' in src_dict and 'question' in cache_dict:
            if src_dict['question'].lower() == cache_dict['question'].lower():
                return self.range()[1]
            if 'embedding' not in src_dict or 'embedding' not in cache_dict or src_dict['embedding'] is None or cache_dict['embedding'] is None:
                assert self.question_encoder, 'You need to a valid question_embedding_function to generate question embedding in the evaluator.'
                src_dict['embedding'] = self.question_encoder(src_dict['question'])
                cache_dict['embedding'] = self.question_encoder(cache_dict['question'])
        src_embedding = (
            self.normalize(src_dict['embedding'])
            if self.enable_normal
            else src_dict['embedding']
        )
        cache_embedding = cache_dict['embedding']
        cache_embedding = (
            self.normalize(cache_embedding) if self.enable_normal else cache_embedding
        )
        return self.range()[1] - np.linalg.norm(src_embedding - cache_embedding)

    def range(self) -> Tuple[float, float]:
        """Range of similarity score.

        :return: minimum and maximum of similarity score.
        """
        return 0.0, 2.0



from typing import Optional, Callable, List

from gptcache.utils.error import CacheError


class Config:
    """Pass configuration.

    :param log_time_func: optional, customized log time function
    :type log_time_func: Optional[Callable[[str, float], None]]
    :param similarity_threshold: a threshold ranged from 0 to 1 to filter search results with similarity score higher \
     than the threshold. When it is 0, there is no hits. When it is 1, all search results will be returned as hits.
    :type similarity_threshold: float
    :param prompts: optional, if the request content will remove the prompt string when the request contains the prompt list
    :type prompts: Optional[List[str]]
    :param template: optional, if the request content will remove the template string and only keep the parameter value in the template
    :type template: Optional[str]
    :param auto_flush: it will be automatically flushed every time xx pieces of data are added, default to 20
    :type auto_flush: int
    :param enable_token_counter: enable token counter, default to False
    :type enable_token_counter: bool
    :param input_summary_len: optional, summarize input to specified length.
    :type input_summary_len: Optional[int]
    :param skip_list: for sequence preprocessing, skip those sentences in skip_list.
    :type skip_list: Optional[List[str]]
    :param context_len: optional, the length of context.
    :type context_len: Optional[int]

    Example:
        .. code-block:: python

            from gptcache import Config

            configs = Config(similarity_threshold=0.6)
    """

    def __init__(
            self,
            log_time_func: Optional[Callable[[str, float], None]] = None,
            similarity_threshold: float = 0.8,
            prompts: Optional[List[str]] = None,
            template: Optional[str] = None,
            auto_flush: int = 20,
            enable_token_counter: bool = True,
            input_summary_len: Optional[int] = None,
            context_len: Optional[int] = None,
            skip_list: List[str] = None,
            data_check: bool = False,
            disable_report: bool = False,
    ):
        if similarity_threshold < 0 or similarity_threshold > 1:
            raise CacheError(
                "Invalid the similarity threshold param, reasonable range: 0-1"
            )
        self.log_time_func = log_time_func
        self.similarity_threshold = similarity_threshold
        self.prompts = prompts
        self.template = template
        self.auto_flush = auto_flush
        self.enable_token_counter = enable_token_counter
        self.input_summary_len = input_summary_len
        self.context_len = context_len
        if skip_list is None:
            skip_list = ["system", "assistant"]
        self.skip_list = skip_list
        self.data_check = data_check
        self.disable_report = disable_report



from typing import Tuple, Dict, Any

from gptcache.similarity_evaluation import SimilarityEvaluation
from gptcache.utils import import_cohere

import_cohere()

import cohere  # pylint: disable=C0413


class CohereRerank(SimilarityEvaluation):
    """Use the Cohere Rerank API to evaluate relevance of question and answer.

    Reference: https://docs.cohere.com/reference/rerank-1

    :param model: model name, defaults to 'rerank-english-v2.0', and multilingual option: rerank-multilingual-v2.0.
    :type model: str
    :param api_key: cohere api key, defaults to None.
    :type api_key: str

    Example:
        .. code-block:: python

            from gptcache.similarity_evaluation import CohereRerankEvaluation

            evaluation = CohereRerankEvaluation()
            score = evaluation.evaluation(
                {
                    'question': 'What is the color of sky?'
                },
                {
                    'answer': 'the color of sky is blue'
                }
            )
    """

    def __init__(self, model: str = "rerank-english-v2.0", api_key: str = None):
        self.co = cohere.Client(api_key)
        self.model = model

    def evaluation(self, src_dict: Dict[str, Any], cache_dict: Dict[str, Any], **kwargs) -> float:
        response = self.co.rerank(
            model=self.model,
            query=src_dict["question"],
            documents=cache_dict["answer"],
            top_n=1,
        )
        if len(response.results) == 0:
            return 0
        return response.results[0].relevance_score

    def range(self) -> Tuple[float, float]:
        return 0.0, 1.0



import asyncio
import json

from gptcache.utils import import_httpx

import_httpx()

import httpx  # pylint: disable=C0413


_CLIENT_HEADER = {"Content-Type": "application/json", "Accept": "application/json"}


class Client:
    """GPTCache client to send requests to GPTCache server.

    :param uri: the uri leads to the server, defaults to "http://localhost:8000".
    :type uri: str

    Example:
        .. code-block:: python

            from gptcache import client

            client = Client(uri="http://localhost:8000")
            client.put("Hi", "Hi back")
            ans = client.get("Hi")
    """

    def __init__(self, uri: str = "http://localhost:8000"):
        self._uri = uri

    async def _put(self, question: str, answer: str):
        async with httpx.AsyncClient() as client:
            data = {
                "prompt": question,
                "answer": answer,
            }

            response = await client.post(
                f"{self._uri}/put", headers=_CLIENT_HEADER, data=json.dumps(data)
            )

        return response.status_code

    async def _get(self, question: str):
        async with httpx.AsyncClient() as client:
            data = {
                "prompt": question,
            }

            response = await client.post(
                f"{self._uri}/get", headers=_CLIENT_HEADER, data=json.dumps(data)
            )

        return response.json().get("answer")

    def put(self, question: str, answer: str):
        """
        :param question: the question to be put.
        :type question: str
        :param answer: the answer to the question to be put.
        :type answer: str
        :return: status code.
        """
        return asyncio.run(self._put(question, answer))

    def get(self, question: str):
        """
        :param question: the question to get an answer.
        :type question: str
        :return: answer to the question.
        """
        return asyncio.run(self._get(question))



from typing import List

import numpy as np

from gptcache.manager.vector_data.base import VectorBase, VectorData
from gptcache.utils import import_sqlalchemy

import_sqlalchemy()

# pylint: disable=wrong-import-position
from sqlalchemy import create_engine, Column, Index, text  # pylint: disable=C0413
from sqlalchemy.types import (  # pylint: disable=C0413
    Integer,
    UserDefinedType
)
from sqlalchemy.orm import sessionmaker  # pylint: disable=C0413
from sqlalchemy.ext.declarative import declarative_base  # pylint: disable=C0413

Base = declarative_base()


class _VectorType(UserDefinedType):
    """
    pgvector type mapping for sqlalchemy
    """
    cache_ok = True

    def __init__(self, precision=8):
        self.precision = precision

    def get_col_spec(self, **_):
        return f"vector({self.precision})"

    # pylint: disable=unused-argument
    def bind_processor(self, dialect):
        return lambda value: value

    # pylint: disable=unused-argument
    def result_processor(self, dialect, coltype):
        return lambda value: value


def _get_model_and_index(table_prefix, vector_dimension, index_type, lists):
    class VectorStoreTable(Base):
        """
        vector store table
        """

        __tablename__ = table_prefix + "_pg_vector_store"
        __table_args__ = {"extend_existing": True}
        id = Column(Integer, primary_key=True, autoincrement=False)
        embedding = Column(_VectorType(vector_dimension), nullable=False)

    vector_store_index = Index(
        f"idx_{table_prefix}_pg_vector_store_embedding",
        text(f"embedding {index_type}"),
        postgresql_using="ivfflat",
        postgresql_with={"lists": lists}
    )

    vector_store_index.table = VectorStoreTable.__table__

    return VectorStoreTable, vector_store_index


class PGVector(VectorBase):
    """vector store: pgvector

    :param url: the connection url for PostgreSQL database, defaults to 'postgresql://postgres@localhost:5432/postgres'.
    :type url: str
    :type collection_name: str
    :param dimension: the dimension of the vector, defaults to 0.
    :type dimension: int
    :param top_k: the number of the vectors results to return, defaults to 1.
    :type top_k: int
    :param index_params: the index parameters for pgvector, defaults to 'vector_l2_ops' index:
                         {"index_type": "L2", "params": {"lists": 100, "probes": 10}.
    :type index_params: dict
    """

    INDEX_PARAM = {
        "L2": {"operator": "<->", "name": "vector_l2_ops"},  # The only one supported now
        "cosine": {"operator": "<=>", "name": "vector_cosine_ops"},
        "inner_product": {"operator": "<->", "name": "vector_ip_ops"},
    }

    def __init__(
            self,
            url: str,
            index_params: dict,
            collection_name: str = "gptcache",
            dimension: int = 0,
            top_k: int = 1,
    ):
        if dimension <= 0:
            raise ValueError(
                f"invalid `dim` param: {dimension} in the pgvector store."
            )
        self.dimension = dimension
        self.top_k = top_k
        self.index_params = index_params
        self._url = url
        self._store, self._index = _get_model_and_index(
            collection_name,
            dimension,
            index_type=self.INDEX_PARAM[index_params["index_type"]]["name"],
            lists=index_params["params"]["lists"]
        )
        self._connect(url)
        self._create_collection()

    def _connect(self, url):
        self._engine = create_engine(url, echo=False)
        self._session = sessionmaker(bind=self._engine)  # pylint: disable=invalid-name

    def _create_collection(self):
        with self._engine.connect() as con:
            con.execution_options(isolation_level="AUTOCOMMIT").execute(text("CREATE EXTENSION IF NOT EXISTS vector;"))

        self._store.__table__.create(bind=self._engine, checkfirst=True)
        self._index.create(bind=self._engine, checkfirst=True)

    def _query(self, session):
        return session.query(self._store)

    def _format_data_for_search(self, data):
        return f"[{','.join(map(str, data))}]"

    def mul_add(self, datas: List[VectorData]):
        data_array, id_array = map(list, zip(*((data.data, data.id) for data in datas)))
        np_data = np.array(data_array).astype("float32")
        entities = [{"id": id, "embedding": embedding.tolist()} for id, embedding in zip(id_array, np_data)]

        with self._session() as session:
            session.bulk_insert_mappings(self._store, entities)
            session.commit()

    def search(self, data: np.ndarray, top_k: int = -1):
        if top_k == -1:
            top_k = self.top_k

        formatted_data = self._format_data_for_search(data.reshape(1, -1)[0].tolist())
        index_config = self.INDEX_PARAM[self.index_params["index_type"]]
        similarity = self._store.embedding.op(index_config["operator"])(formatted_data)
        with self._session() as session:
            session.execute(text(f"SET LOCAL ivfflat.probes = {self.index_params['params']['probes'] or 10};"))
            search_result = self._query(session).add_columns(
                similarity.label("distances")
            ).order_by(
                similarity
            ).limit(top_k).all()
            search_result = [(r[1], r[0].id) for r in search_result]

        return search_result

    def delete(self, ids):
        with self._session() as session:
            self._query(session).filter(self._store.id.in_(ids)).delete()
            session.commit()

    def rebuild(self, ids=None):  # pylint: disable=unused-argument
        with self._engine.connect() as con:
            con.execution_options(isolation_level="AUTOCOMMIT").execute(
                text(f"REINDEX INDEX CONCURRENTLY {self._index.name}"))

    def flush(self):
        with self._session() as session:
            session.flush()

    def close(self):
        self.flush()



from typing import Dict, Tuple, Any
from gptcache.utils import import_sbert
from gptcache.similarity_evaluation import SimilarityEvaluation
import_sbert()
from sentence_transformers import CrossEncoder # pylint: disable=C0413

class SbertCrossencoderEvaluation(SimilarityEvaluation):
    """Using SBERT crossencoders to evaluate sentences pair similarity.

    This evaluator use the crossencoder model to evaluate the similarity of two sentences.

    :param model: model name of SbertCrossencoderEvaluation. Default is 'cross-encoder/quora-distilroberta-base'.
    Check more please refer to https://www.sbert.net/docs/pretrained_cross-encoders.html#quora-duplicate-questions.
    :type model: str

    Example:
        .. code-block:: python

            from gptcache.similarity_evaluation import SbertCrossencoderEvaluation

            evaluation = SbertCrossencoderEvaluation()
            score = evaluation.evaluation(
                {
                    'question': 'What is the color of sky?'
                },
                {
                    'question': 'hello'
                }
            )
    """
    def __init__(self, model: str="cross-encoder/quora-distilroberta-base"):
        self.model = CrossEncoder(model)

    def evaluation(
        self, src_dict: Dict[str, Any], cache_dict: Dict[str, Any], **_
    ) -> float:
        """Evaluate the similarity score of pair.

        :param src_dict: the query dictionary to evaluate with cache.
        :type src_dict: Dict
        :param cache_dict: the cache dictionary.
        :type cache_dict: Dict

        :return: evaluation score.
        """
        try:
            src_question = src_dict["question"]
            cache_question = cache_dict["question"]
            if src_question.lower() == cache_question.lower():
                return 1
            return self.model.predict([(src_question, cache_question)])[0]
        except Exception: # pylint: disable=W0703
            return 0

    def range(self) -> Tuple[float, float]:
        """Range of similarity score.

        :return: minimum and maximum of similarity score.
        """
        return 0.0, 1.0



from typing import Dict, Any

from gptcache.similarity_evaluation.similarity_evaluation import SimilarityEvaluation

__all__ = [
    "SimilarityEvaluation",
    "OnnxModelEvaluation",
    "NumpyNormEvaluation",
    "SearchDistanceEvaluation",
    "ExactMatchEvaluation",
    "KReciprocalEvaluation",
    "CohereRerankEvaluation",
    "SequenceMatchEvaluation",
    "TimeEvaluation",
    "SbertCrossencoderEvaluation"
]

from gptcache.utils.lazy_import import LazyImport

onnx = LazyImport("onnx", globals(), "gptcache.similarity_evaluation.onnx")
np = LazyImport("np", globals(), "gptcache.similarity_evaluation.np")
distance = LazyImport("simple", globals(), "gptcache.similarity_evaluation.distance")
exact_match = LazyImport(
    "exact_match", globals(), "gptcache.similarity_evaluation.exact_match"
)
kreciprocal = LazyImport(
    "kreciprocal", globals(), "gptcache.similarity_evaluation.kreciprocal"
)
cohere = LazyImport(
    "cohere", globals(), "gptcache.similarity_evaluation.cohere_rerank"
)
sequence_match = LazyImport(
    "sequence_match", globals(), "gptcache.similarity_evaluation.sequence_match"
)
time = LazyImport(
    "time", globals(), "gptcache.similarity_evaluation.time"
)

time = LazyImport(
    "time", globals(), "gptcache.similarity_evaluation.time"
)

sbert_crossencoder = LazyImport(
    "sbert_crossencoder", globals(), "gptcache.similarity_evaluation.sbert_crossencoder"
)

def OnnxModelEvaluation(model="GPTCache/albert-duplicate-onnx"):
    return onnx.OnnxModelEvaluation(model)


def NumpyNormEvaluation(enable_normal: bool = False, **kwargs):
    return np.NumpyNormEvaluation(enable_normal, **kwargs)


def SearchDistanceEvaluation(max_distance=4.0, positive=False):
    return distance.SearchDistanceEvaluation(max_distance, positive)


def ExactMatchEvaluation():
    return exact_match.ExactMatchEvaluation()


def KReciprocalEvaluation(vectordb, top_k=3, max_distance=4.0, positive=False):
    return kreciprocal.KReciprocalEvaluation(vectordb, top_k, max_distance, positive)


def CohereRerankEvaluation(model: str = "rerank-english-v2.0", api_key: str = None):
    return cohere.CohereRerank(model=model, api_key=api_key)


def SequenceMatchEvaluation(weights, embedding_extractor, embedding_config: Dict[str, Any] = None):
    return sequence_match.SequenceMatchEvaluation(weights, embedding_extractor, embedding_config=embedding_config)


def TimeEvaluation(evaluation: str, evaluation_config: Dict[str, Any], time_range: float = 86400.0):
    return time.TimeEvaluation(evaluation, evaluation_config=evaluation_config, time_range=time_range)

def SbertCrossencoderEvaluation(model: str = "cross-encoder/quora-distilroberta-base"):
    return sbert_crossencoder.SbertCrossencoderEvaluation(model)



from typing import List

import numpy as np

from gptcache.manager.vector_data.base import VectorBase, VectorData
from gptcache.utils import import_chromadb, import_torch

import_torch()
import_chromadb()

import chromadb  # pylint: disable=C0413


class Chromadb(VectorBase):
    """vector store: Chromadb

    :param client_settings: the setting for Chromadb.
    :type client_settings: Settings
    :param persist_directory: the directory to persist, defaults to .chromadb/ in the current directory.
    :type persist_directory: str
    :param collection_name: the name of the collection in Chromadb, defaults to 'gptcache'.
    :type collection_name: str
    :param top_k: the number of the vectors results to return, defaults to 1.
    :type top_k: int

    """

    def __init__(
        self,
        client_settings=None,
        persist_directory=None,
        collection_name: str = "gptcache",
        top_k: int = 1,
    ):
        self.top_k = top_k

        if client_settings:
            self._client_settings = client_settings
        else:
            self._client_settings = chromadb.config.Settings()
            if persist_directory is not None:
                self._client_settings = chromadb.config.Settings(
                    chroma_db_impl="duckdb+parquet", persist_directory=persist_directory
                )
        self._client = chromadb.Client(self._client_settings)
        self._persist_directory = persist_directory
        self._collection = self._client.get_or_create_collection(name=collection_name)

    def mul_add(self, datas: List[VectorData]):
        data_array, id_array = map(list, zip(*((data.data.tolist(), str(data.id)) for data in datas)))
        self._collection.add(embeddings=data_array, ids=id_array)

    def search(self, data, top_k: int = -1):
        if self._collection.count() == 0:
            return []
        if top_k == -1:
            top_k = self.top_k
        results = self._collection.query(
            query_embeddings=[data.tolist()],
            n_results=top_k,
            include=["distances"],
        )
        return list(zip(results["distances"][0], [int(x) for x in results["ids"][0]]))

    def delete(self, ids):
        self._collection.delete([str(x) for x in ids])

    def rebuild(self, ids=None):  # pylint: disable=unused-argument
        return True

    def get_embeddings(self, data_id: str):
        vec_emb = self._collection.get(
                str(data_id),
                include=["embeddings"],
            )["embeddings"]
        if vec_emb is None or len(vec_emb) < 1:
            return None
        vec_emb = np.asarray(vec_emb[0], dtype="float32")
        return vec_emb

    def update_embeddings(self, data_id: str, emb: np.ndarray):
        self._collection.update(
            ids=str(data_id),
            embeddings=emb.tolist(),
        )



from abc import ABCMeta, abstractmethod
from typing import Tuple, Dict, Any


class SimilarityEvaluation(metaclass=ABCMeta):
    """Similarity Evaluation interface,
    determine the similarity between the input request and the requests from the Vector Store.
    Based on this similarity, it determines whether a request matches the cache.

    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.similarity_evaluation import SearchDistanceEvaluation

            cache.init(
                similarity_evaluation=SearchDistanceEvaluation()
            )
    """

    @abstractmethod
    def evaluation(
        self, src_dict: Dict[str, Any], cache_dict: Dict[str, Any], **kwargs
    ) -> float:
        """Evaluate the similarity score of the user and cache requests pair.

        :param src_dict: the user request params.
        :type src_dict: Dict
        :param cache_dict: the cache request params.
        :type cache_dict: Dict
        """
        pass

    @abstractmethod
    def range(self) -> Tuple[float, float]:
        """Range of similarity score.

        :return: the range of similarity score, which is the min and max values
        :rtype: Tuple[float, float]
        """
        pass



from typing import Dict, Any

import numpy as np

from gptcache.processor import ContextProcess
from gptcache.utils import import_huggingface

import_huggingface()

import transformers  # pylint: disable=C0413


def summarize_to_length(summarizer, text, target_len, max_len=1024):
    tokenizer = summarizer.tokenizer

    def token_length(text):
        return len(tokenizer.encode(text))

    segment_len = max_len - 100
    summary_result = text
    while token_length(text) > target_len:
        tokens = tokenizer.encode(text)
        segments = [
            tokens[i : i + segment_len] for i in range(0, len(tokens), segment_len - 1)
        ]
        summary_result = ""
        for segment in segments:
            len_seg = int(len(segment) / 4)
            summary = summarizer(
                tokenizer.decode(segment),
                min_length=max(len_seg - 10, 1),
                max_length=len_seg,
            )
            summary_result += summary[0]["summary_text"]
        text = summary_result
    return summary_result


class SummarizationContextProcess(ContextProcess):
    """A context processor for summarizing large amounts of text data using a summarizer model.

    :param summarizer: The summarizer model to use for summarization.
    :type summarizer: transformers.PreTrainedModel
    :param tokenizer: The tokenizer to use for tokenizing the text data.
    It used for measuring the output length.
    :type tokenizer: transformers.PreTrainedTokenizer
    :param target_length: The length of the summarized text.
    :type target_length: int

    Example:
        .. code-block:: python

            from gptcache.processor.context.summarization_context import SummarizationContextProcess

            context_process = SummarizationContextProcess()
            cache.init(pre_embedding_func=context_process.pre_process)
    """

    def __init__(
        self, model_name="facebook/bart-large-cnn", tokenizer=None, target_length=512
    ):
        summarizer = transformers.pipeline(task="summarization", model=model_name)
        self.summarizer = summarizer
        self.target_length = target_length
        if tokenizer is None:
            tokenizer = transformers.RobertaTokenizer.from_pretrained("roberta-base")
        self.tokenizer = tokenizer
        self.content = ""

    def summarize_to_sentence(self, sentences, target_size=1000):
        lengths = []
        for sentence in sentences:
            lengths.append(len(sentence))
        total_length = np.array(lengths).sum()
        target_lengths = [int(target_size * l / total_length) for l in lengths]
        target_sentences = []
        for sent, target_len in zip(sentences, target_lengths):
            if len(self.tokenizer.tokenize(sent)) > target_len:
                response = summarize_to_length(
                    self.summarizer, sent, target_len, self.tokenizer.model_max_length
                )
                target_sentence = response
            else:
                target_sentence = sent
            target_sentences.append(target_sentence)
        result = ""
        for target_sentence in target_sentences:
            result = result + target_sentence
        return result

    def format_all_content(self, data: Dict[str, Any], **params: Dict[str, Any]):
        contents = []
        for query in data["messages"]:
            contents.append(query)
        self.content = contents

    def process_all_content(self) -> (Any, Any):
        def serialize_content(content):
            ret = ""
            for message in content:
                ret += "[#RS]{}[#RE][#CS]{}[#CE]".format(
                    message["role"], message["content"]
                )
            return ret

        result = self.summarize_to_sentence(
            [message["content"] for message in self.content], self.target_length
        )
        save_content = serialize_content(self.content)
        embedding_content = result
        return save_content, embedding_content



class Report:
    """Get GPTCache report including time and counts for different operations."""

    def __init__(self):
        self.op_pre = OpCounter()
        self.op_embedding = OpCounter()
        self.op_search = OpCounter()
        self.op_data = OpCounter()
        self.op_evaluation = OpCounter()
        self.op_post = OpCounter()
        self.op_llm = OpCounter()
        self.op_save = OpCounter()
        self.hint_cache_count = 0

    def pre(self, delta_time):
        """Pre-process counts and time.

        :param delta_time: additional runtime.
        """
        self.op_pre.total_time += delta_time
        self.op_pre.count += 1

    def embedding(self, delta_time):
        """Embedding counts and time.

        :param delta_time: additional runtime.
        """
        self.op_embedding.total_time += delta_time
        self.op_embedding.count += 1

    def search(self, delta_time):
        """Search counts and time.

        :param delta_time: additional runtime.
        """
        self.op_search.total_time += delta_time
        self.op_search.count += 1

    def data(self, delta_time):
        """Get data counts and time.

        :param delta_time: additional runtime.
        """

        self.op_data.total_time += delta_time
        self.op_data.count += 1

    def evaluation(self, delta_time):
        """Evaluation counts and time.

        :param delta_time: additional runtime.
        """
        self.op_evaluation.total_time += delta_time
        self.op_evaluation.count += 1

    def post(self, delta_time):
        """Post-process counts and time.

        :param delta_time: additional runtime.
        """
        self.op_post.total_time += delta_time
        self.op_post.count += 1

    def llm(self, delta_time):
        """LLM counts and time.

        :param delta_time: additional runtime.
        """
        self.op_llm.total_time += delta_time
        self.op_llm.count += 1

    def save(self, delta_time):
        """Save counts and time.

        :param delta_time: additional runtime.
        """
        self.op_save.total_time += delta_time
        self.op_save.count += 1

    def average_pre_time(self):
        """Average pre-process time."""
        return self.op_pre.average()

    def average_embedding_time(self):
        """Average embedding time."""
        return self.op_embedding.average()

    def average_search_time(self):
        """Average search time."""
        return self.op_search.average()

    def average_data_time(self):
        """Average data time."""
        return self.op_data.average()

    def average_evaluation_time(self):
        """Average evaluation time."""
        return self.op_evaluation.average()

    def average_post_time(self):
        """Average post-process time."""
        return self.op_post.average()

    def average_llm_time(self):
        """Average LLM time."""
        return self.op_llm.average()

    def average_save_time(self):
        """Average save time."""
        return self.op_save.average()

    def hint_cache(self):
        """hint cache count."""
        self.hint_cache_count += 1


class OpCounter:
    """Operation counter."""

    count = 0
    """Operation count."""
    total_time = 0
    """Total time."""

    def average(self):
        """Average time."""
        return round(self.total_time / self.count, 4) if self.count != 0 else 0



import time

from gptcache import cache


def time_cal(func, func_name=None, report_func=None):
    def inner(*args, **kwargs):
        time_start = time.time()
        res = func(*args, **kwargs)
        delta_time = time.time() - time_start
        if cache.config.log_time_func:
            cache.config.log_time_func(
                func.__name__ if func_name is None else func_name, delta_time
            )
        if report_func is not None:
            report_func(delta_time)
        return res

    return inner



from typing import Tuple, Dict, Any

from gptcache.similarity_evaluation import SimilarityEvaluation


class SearchDistanceEvaluation(SimilarityEvaluation):
    """Using search distance to evaluate sentences pair similarity.

    This is the evaluator to compare two embeddings according to their distance computed in embedding retrieval stage.
    In the retrieval stage, `search_result` is the distance used for approximate nearest neighbor search and have been
    put into `cache_dict`. `max_distance` is used to bound this distance to make it between [0-`max_distance`]. `positive` is
    used to indicate this distance is directly proportional to the similarity of two entites. If `positive` is set `False`,
    `max_distance` will be used to substract this distance to get the final score.

    :param max_distance: the bound of maximum distance.
    :type max_distance: float
    :param positive: if the larger distance indicates more similar of two entities, It is True. Otherwise it is False.
    :type positive: bool

    Example:
        .. code-block:: python

            from gptcache.similarity_evaluation import SearchDistanceEvaluation

            evaluation = SearchDistanceEvaluation()
            score = evaluation.evaluation(
                {},
                {
                    "search_result": (1, None)
                }
            )
    """

    def __init__(self, max_distance=4.0, positive=False):
        self.max_distance = max_distance
        self.positive = positive

    def evaluation(
        self, src_dict: Dict[str, Any], cache_dict: Dict[str, Any], **_
    ) -> float:
        """Evaluate the similarity score of pair.
        :param src_dict: the query dictionary to evaluate with cache.
        :type src_dict: Dict
        :param cache_dict: the cache dictionary.
        :type cache_dict: Dict

        :return: evaluation score.
        """
        distance, _ = cache_dict["search_result"]
        if distance < 0:
            distance = 0
        elif distance > self.max_distance:
            distance = self.max_distance
        if self.positive:
            return distance
        return self.max_distance - distance

    def range(self) -> Tuple[float, float]:
        """Range of similarity score.

        :return: minimum and maximum of similarity score.
        """
        return 0.0, self.max_distance



from datetime import datetime
from typing import List, Optional

import numpy as np

from gptcache.manager.scalar_data.base import (
    CacheStorage,
    CacheData,
    Question,
    QuestionDep,
)
from gptcache.utils import import_mongodb

import_mongodb()

# pylint: disable=C0413
from mongoengine import Document
from mongoengine import fields
import mongoengine as me


def get_models():
    class Questions(Document):
        """
        questions collection
        """

        meta = {"collection": "questions", "indexes": ["deleted"]}
        _id = fields.SequenceField()
        question = fields.StringField()
        create_on = fields.DateTimeField(default=datetime.now())
        last_access = fields.DateTimeField(default=datetime.now())
        embedding_data = fields.BinaryField()
        deleted = fields.IntField(default=0)

        @property
        def oid(self):
            return self._id

    class Answers(Document):
        """
        answer collection
        """

        _id = fields.SequenceField()
        meta = {"collection": "answers", "indexes": ["question_id"]}
        answer = fields.StringField()
        answer_type = fields.IntField()
        question_id = fields.IntField()

        @property
        def oid(self):
            return self._id

    class Sessions(Document):
        """
        session collection
        """

        meta = {"collection": "sessions", "indexes": ["question_id"]}
        _id = fields.SequenceField()
        session_id = fields.StringField()
        session_question = fields.StringField()
        question_id = fields.IntField()

        @property
        def oid(self):
            return self._id

    class QuestionDeps(Document):
        """
        Question Dep collection
        """

        meta = {"collection": "question_deps", "indexes": ["question_id"]}
        _id = fields.SequenceField()
        question_id = fields.IntField()
        dep_name = fields.StringField()
        dep_data = fields.StringField()
        dep_type = fields.IntField()

        @property
        def oid(self):
            return self._id

    class Report(Document):
        """
        Report
        """

        meta = {
            "collection": "report",
            "indexes": ["cache_question_id", "similarity", "cache_delta_time"],
        }
        _id = fields.SequenceField()
        user_question = fields.StringField()
        cache_question_id = fields.IntField()
        cache_question = fields.StringField()
        cache_answer = fields.StringField()
        similarity = fields.FloatField()
        cache_delta_time = fields.FloatField()
        cache_time = fields.DateTimeField(default=datetime.now())
        extra = fields.StringField()

        @property
        def oid(self):
            return self._id

    return Questions, Answers, QuestionDeps, Sessions, Report


class MongoStorage(CacheStorage):
    """
    Using mongoengine as ORM to manage mongodb documents.
    By default, data is stored 'gptcache' database and following collections are created to store the data
        1. 'sessions'
        2. 'answers'
        3. 'questions'
        4. 'question_deps'

    :param host: mongodb host, default value 'localhost'
    :type host: str
    :param port: mongodb port, default value 27017
    :type host: int
    :param dbname: database name, default value 'gptcache'
    :type host: str
    :param : Mongo database name, default value 'gptcache'
    :type host: str
    :param username: username for authentication, default value None
    :type host: str
    :param password: password for authentication, default value None
    :type host: str

    Example:
        .. code-block:: python

            from gptcache.manager import CacheBase, manager_factory

            cache_store = CacheBase('mongo',
                mongo_host="localhost",
                mongo_port=27017,
                dbname="gptcache",
                username=None,
                password=None,
            )
            # or
            data_manager = manager_factory("mongo,faiss", data_dir="./workspace",
                scalar_params={
                    "mongo_host": "localhost",
                    "mongo_port": 27017,
                    "dbname"="gptcache",
                    "username"="",
                    "password"="",
                },
                vector_params={"dimension": 128},
            )
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 27017,
        dbname: str = "gptcache",
        username: str = None,
        password: str = None,
        **kwargs
    ):
        self.con = me.connect(
            host=host,
            port=port,
            db=dbname,
            username=username,
            password=password,
            **kwargs
        )
        (
            self._ques,
            self._answer,
            self._ques_dep,
            self._session,
            self._report,
        ) = get_models()

    def create(self):
        pass

    def _insert(self, data: CacheData):
        ques_data = self._ques(
            question=data.question
            if isinstance(data.question, str)
            else data.question.content,
            embedding_data=data.embedding_data.tobytes()
            if data.embedding_data is not None
            else None,
        )
        ques_data.save()
        if isinstance(data.question, Question) and data.question.deps is not None:
            all_deps = []
            for dep in data.question.deps:
                all_deps.append(
                    self._ques_dep(
                        question_id=ques_data.oid,
                        dep_name=dep.name,
                        dep_data=dep.data,
                        dep_type=dep.dep_type,
                    )
                )
            self._ques_dep.objects.insert(all_deps)

        answers = data.answers if isinstance(data.answers, list) else [data.answers]
        all_data = []
        for answer in answers:
            answer_data = self._answer(
                question_id=ques_data.oid,
                answer=answer.answer,
                answer_type=int(answer.answer_type),
            )
            all_data.append(answer_data)
        self._answer.objects.insert(all_data)

        if data.session_id:
            session_data = self._session(
                question_id=ques_data.oid,
                session_id=data.session_id,
                session_question=data.question
                if isinstance(data.question, str)
                else data.question.content,
            )
            self._session.objects.insert(session_data)

        return ques_data.oid

    def batch_insert(self, all_data: List[CacheData]):
        ids = []
        for data in all_data:
            ids.append(self._insert(data))
        return ids

    def get_data_by_id(self, key) -> Optional[CacheData]:
        qs = self._ques.objects.get(_id=key, deleted=0)
        if qs is None:
            return None
        last_access = qs.last_access
        qs.last_access = datetime.now()
        qs.save()
        answers = self._answer.objects(question_id=qs.oid)
        deps = self._ques_dep.objects(question_id=qs.oid)
        session_ids = self._session.objects(question_id=qs.oid)

        res_ans = [(item.answer, item.answer_type) for item in answers]
        res_deps = [
            QuestionDep(item.dep_name, item.dep_data, item.dep_type) for item in deps
        ]
        return CacheData(
            question=qs.question if not deps else Question(qs.question, res_deps),
            answers=res_ans,
            embedding_data=np.frombuffer(qs.embedding_data, dtype=np.float32),
            session_id=session_ids,
            create_on=qs.create_on,
            last_access=last_access,
        )

    def mark_deleted(self, keys):
        self._ques.objects(_id__in=keys).update(deleted=-1)

    def clear_deleted_data(self):
        questions = self._ques.objects(deleted=-1).only("_id")
        q_ids = [obj.oid for obj in questions]
        self._answer.objects(question_id__in=q_ids).delete()
        self._ques_dep.objects(question_id__in=q_ids).delete()
        self._session.objects(question_id__in=q_ids).delete()
        questions.delete()

    def get_ids(self, deleted: bool = True):
        state = -1 if deleted else 0
        res = [obj.oid for obj in self._ques.objects(deleted=state).only("_id")]
        return res

    def count(self, state: int = 0, is_all: bool = False):
        if is_all:
            return self._ques.objects.count()
        return self._ques.objects(deleted=state).count()

    def add_session(self, question_id, session_id, session_question):
        self._session(
            question_id=question_id,
            session_id=session_id,
            session_question=session_question,
        ).save()

    def list_sessions(self, session_id=None, key=None):
        query = {}
        if session_id:
            query["session_id"] = session_id
        if key:
            query["question_id"] = key

        return self._session.objects(__raw__=query)

    def delete_session(self, keys):
        self._session.objects(question_id__in=keys).delete()

    def count_answers(self):
        return self._answer.objects.count()

    def report_cache(
        self,
        user_question,
        cache_question,
        cache_question_id,
        cache_answer,
        similarity_value,
        cache_delta_time,
    ):
        report_data = self._report(
            user_question=user_question,
            cache_question=cache_question,
            cache_question_id=cache_question_id,
            cache_answer=cache_answer,
            similarity=similarity_value,
            cache_delta_time=cache_delta_time,
        )
        report_data.save()

    def close(self):
        me.disconnect()
        self.con.close()



from abc import ABCMeta
from typing import Any, Dict, Callable, Optional


class BaseCacheLLM(metaclass=ABCMeta):
    """Base LLM, When you have enhanced llm without using the original llm api,
    you can use this class as a proxy to use the ability of the cache.

    NOTE: Please make sure that the custom llm returns the same value as the original llm.

    For example, if you use the openai proxy, you perform delay statistics before sending the openai request,
    and then you package this part of the function, so you may have a separate package, which is different from openai.
    If the api request parameters and return results you wrap are the same as the original ones,
    then you can use this class to obtain cache-related capabilities.

    Example:
        .. code-block:: python

            import time

            import openai

            from gptcache import Cache
            from gptcache.adapter import openai as cache_openai


            def proxy_openai_chat_complete(*args, **kwargs):
                start_time = time.time()
                res = openai.ChatCompletion.create(*args, **kwargs)
                print("Consume Time Spent =", round((time.time() - start_time), 2))
                return res


            llm_cache = Cache()

            cache_openai.ChatCompletion.llm = proxy_openai_chat_complete
            cache_openai.ChatCompletion.cache_args = {"cache_obj": llm_cache}

            cache_openai.ChatCompletion.create(
                model="gpt-3.5-turbo",
                messages=[
                    {
                        "role": "user",
                        "content": "What's GitHub?",
                    }
                ],
            )
    """

    llm: Optional[Callable] = None
    """
    On a cache miss, if that variable is set, it will be called;
    if not, it will call the original llm.
    """

    cache_args: Dict[str, Any] = {}
    """
    It can be used to set some cache-related public parameters.
    If you don't want to set the same parameters every time when using cache, say cache_obj, you can use it.
    """

    @classmethod
    def fill_base_args(cls, **kwargs):
        """ Fill the base args to the cache args
        """
        for key, value in cls.cache_args.items():
            if key not in kwargs:
                kwargs[key] = value

        return kwargs



from gptcache.utils import import_sql_client
from gptcache.utils.error import NotFoundError

SQL_URL = {
    "sqlite": "sqlite:///./sqlite.db",
    "duckdb": "duckdb:///./duck.db",
    "postgresql": "postgresql+psycopg2://postgres:123456@127.0.0.1:5432/postgres",
    "mysql": "mysql+pymysql://root:123456@127.0.0.1:3306/mysql",
    "mariadb": "mariadb+pymysql://root:123456@127.0.0.1:3307/mysql",
    "sqlserver": "mssql+pyodbc://sa:Strongpsw_123@127.0.0.1:1434/msdb?driver=ODBC+Driver+17+for+SQL+Server",
    "oracle": "oracle+cx_oracle://oracle:123456@127.0.0.1:1521/?service_name=helowin&encoding=UTF-8&nencoding=UTF-8",
}
TABLE_NAME = "gptcache"


# pylint: disable=import-outside-toplevel
class CacheBase:
    """
    CacheBase to manager the cache storage.

    Generate specific CacheStorage with the configuration. For example, setting for
       `SQLDataBase` (with `name`, `sql_url` and `table_name` params) to manage SQLite, PostgreSQL, MySQL, MariaDB, SQL Server and Oracle.

    :param name: the name of the cache storage, it is support 'sqlite', 'postgresql', 'mysql', 'mariadb', 'sqlserver' and  'oracle' now.
    :type name: str
    :param sql_url: the url of the sql database for cache, such as '<db_type>+<db_driver>://<username>:<password>@<host>:<port>/<database>',
        and the default value is related to the `cache_store` parameter,

        - 'sqlite:///./sqlite.db' for 'sqlite',
        - 'duckdb:///./duck.db' for 'duckdb',
        - 'postgresql+psycopg2://postgres:123456@127.0.0.1:5432/postgres' for 'postgresql',
        - 'mysql+pymysql://root:123456@127.0.0.1:3306/mysql' for 'mysql',
        - 'mariadb+pymysql://root:123456@127.0.0.1:3307/mysql' for 'mariadb',
        - 'mssql+pyodbc://sa:Strongpsw_123@127.0.0.1:1434/msdb?driver=ODBC+Driver+17+for+SQL+Server' for 'sqlserver',
        - 'oracle+cx_oracle://oracle:123456@127.0.0.1:1521/?service_name=helowin&encoding=UTF-8&nencoding=UTF-8' for 'oracle'.
    :type sql_url: str
    :param table_name: the table name for sql database, defaults to 'gptcache'.
    :type table_name: str
    :param table_len_config: the table length config for sql database, defaults to {}. the key includes:

        - 'question_question': the question column size in the question table, default to 3000.
        - 'answer_answer': the answer column size in the answer table, default to 3000.
        - 'session_id': the session id column size in the session table, default to 1000.
        - 'dep_name': the name column size in the dep table, default to 1000.
        - 'dep_data': the data column size in the dep table, default to 3000.
    :type table_len_config: dict

    :return: CacheStorage.

    Example:
        .. code-block:: python

            from gptcache.manager import CacheBase

            cache_base = CacheBase('sqlite')
    """

    def __init__(self):
        raise EnvironmentError(
            "CacheBase is designed to be instantiated, please using the `CacheBase.get(name)`."
        )

    @staticmethod
    def get(name, **kwargs):
        if name in [
            "sqlite",
            "duckdb",
            "postgresql",
            "mysql",
            "mariadb",
            "sqlserver",
            "oracle",
        ]:
            from gptcache.manager.scalar_data.sql_storage import SQLStorage

            sql_url = kwargs.get("sql_url", SQL_URL[name])
            table_name = kwargs.get("table_name", TABLE_NAME)
            table_len_config = kwargs.get("table_len_config", {})
            import_sql_client(name)
            cache_base = SQLStorage(
                db_type=name,
                url=sql_url,
                table_name=table_name,
                table_len_config=table_len_config,
            )
        elif name == "mongo":
            from gptcache.manager.scalar_data.mongo import MongoStorage

            return MongoStorage(
                host=kwargs.get("mongo_host", "localhost"),
                port=kwargs.get("mongo_port", 27017),
                dbname=kwargs.get("dbname", TABLE_NAME),
                username=kwargs.get("username"),
                password=kwargs.get("password")
            )
        elif name == "redis":
            from gptcache.manager.scalar_data.redis_storage import RedisCacheStorage

            return RedisCacheStorage(
                host=kwargs.pop("redis_host", "localhost"),
                port=kwargs.pop("redis_port", 6379),
                global_key_prefix=kwargs.pop("global_key_prefix", TABLE_NAME),
                **kwargs
            )
        elif name == "dynamo":
            from gptcache.manager.scalar_data.dynamo_storage import DynamoStorage

            return DynamoStorage(
                aws_access_key_id=kwargs.get("aws_access_key_id"),
                aws_secret_access_key=kwargs.get("aws_secret_access_key"),
                aws_region_name=kwargs.get("region_name"),
                aws_profile_name=kwargs.get("aws_profile_name"),
                aws_endpoint_url=kwargs.get("endpoint_url"),
            )
        else:
            raise NotFoundError("cache store", name)
        return cache_base



__all__ = ["VectorBase"]

from gptcache.utils.lazy_import import LazyImport

vector_manager = LazyImport(
    "vector_manager", globals(), "gptcache.manager.vector_data.manager"
)


def VectorBase(name: str, **kwargs):
    """Generate specific VectorBase with the configuration.
    """
    return vector_manager.VectorBase.get(name, **kwargs)



# pylint: disable=import-outside-toplevel
from typing import Callable, List, Any

from gptcache.utils.error import NotFoundError


class EvictionBase:
    """
    EvictionBase to evict the cache data.
    """

    def __init__(self):
        raise EnvironmentError(
            "EvictionBase is designed to be instantiated, "
            "please using the `EvictionBase.get(name, policy, maxsize, clean_size)`."
        )

    @staticmethod
    def get(
        name: str,
        policy: str = "LRU",
        maxsize: int = 1000,
        clean_size: int = 0,
        on_evict: Callable[[List[Any]], None] = None,
        **kwargs
    ):
        if not clean_size:
            clean_size = int(maxsize * 0.2)
        if name in "memory":
            from gptcache.manager.eviction.memory_cache import MemoryCacheEviction

            eviction_base = MemoryCacheEviction(
                policy, maxsize, clean_size, on_evict, **kwargs
            )
            return eviction_base
        if name == "redis":
            from gptcache.manager.eviction.redis_eviction import RedisCacheEviction
            if policy == "LRU":
                policy = None
            eviction_base = RedisCacheEviction(policy=policy, **kwargs)
            return eviction_base
        if name == "no_op_eviction":
            from gptcache.manager.eviction.distributed_cache import NoOpEviction
            eviction_base = NoOpEviction()
            return eviction_base

        else:
            raise NotFoundError("eviction base", name)



from typing import Any

from gptcache.adapter.adapter import adapt
from gptcache.manager.scalar_data.base import Answer, DataType
from gptcache.utils import import_huggingface, import_torch

import_torch()
import_huggingface()

from transformers import pipeline  # pylint: disable=wrong-import-position


class Dolly:
    """Wrapper for Dolly (https://github.com/databrickslabs/dolly.git).

    Example using from_model:
        .. code-block:: python

            from gptcache import cache
            from gptcache.processor.pre import get_inputs
            cache.init(pre_embedding_func=get_inputs)

            from gptcache.adapter.dolly import Dolly
            dolly = Dolly.from_model(
                model="databricks/dolly-v2-12b", torch_dtype=torch.bfloat16, trust_remote_code=True, device=0
            )

    Example passing pipeline in directly:
        .. code-block:: python

            import torch
            from transformers import pipeline
            from gptcache import cache
            from gptcache.processor.pre import get_inputs
            cache.init(pre_embedding_func=get_inputs)
            from gptcache.adapter.dolly import Dolly

            pipe = pipeline(
                model="databricks/dolly-v2-12b", torch_dtype=torch.bfloat16, trust_remote_code=True, device=0
            )
            dolly = Dolly(pipe)
    """

    def __init__(self, dolly_pipeline: Any):
        self._dolly_pipeline = dolly_pipeline

    @classmethod
    def from_model(cls, model: str, **kwargs):
        pipe = pipeline(model=model, **kwargs)
        return cls(pipe)

    def __call__(self, prompt: str, **kwargs):
        return adapt(
            self._dolly_pipeline,
            _cache_data_convert,
            _update_cache_callback,
            inputs=prompt,
            **kwargs
        )


def _cache_data_convert(cache_data):
    return [{"generated_text": cache_data, "gptcache": True}]


def _update_cache_callback(llm_data, update_cache_func, *args, **kwargs):  # pylint: disable=unused-argument
    update_cache_func(Answer(llm_data[0]["generated_text"], DataType.STR))
    return llm_data



from datetime import datetime
from typing import List, Optional, Dict

import numpy as np

from gptcache.manager.scalar_data.base import (
    CacheStorage,
    CacheData,
    Question,
    QuestionDep,
)
from gptcache.utils import import_sqlalchemy

import_sqlalchemy()

# pylint: disable=C0413
import sqlalchemy
from sqlalchemy import func, create_engine, Column, Sequence
from sqlalchemy.types import (
    String,
    DateTime,
    LargeBinary,
    Integer,
    Float,
)
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

DEFAULT_LEN_DOCT = {
    "question_question": 3000,
    "answer_answer": 3000,
    "session_id": 1000,
    "dep_name": 1000,
    "dep_data": 3000,
}


def _get_table_len(config: Dict, column_alias: str) -> int:
    if config and column_alias in config and config[column_alias] > 0:
        return config[column_alias]
    return DEFAULT_LEN_DOCT.get(column_alias, 1000)


def get_models(table_prefix, db_type, table_len_config):
    DynamicBase = declarative_base(class_registry={})  # pylint: disable=C0103

    class QuestionTable(DynamicBase):
        """
        question table
        """

        __tablename__ = table_prefix + "_question"
        __table_args__ = {"extend_existing": True}

        if db_type in ("oracle", "duckdb"):
            question_id_seq = Sequence(f"{__tablename__}_id_seq", start=1)
            id = Column(Integer, question_id_seq, primary_key=True, autoincrement=True)
        else:
            id = Column(Integer, primary_key=True, autoincrement=True)
        question = Column(
            String(_get_table_len(table_len_config, "question_question")),
            nullable=False,
        )
        create_on = Column(DateTime, default=datetime.now)
        last_access = Column(DateTime, default=datetime.now)
        embedding_data = Column(LargeBinary, nullable=True)
        deleted = Column(Integer, default=0)

    class AnswerTable(DynamicBase):
        """
        answer table
        """

        __tablename__ = table_prefix + "_answer"
        __table_args__ = {"extend_existing": True}

        if db_type in ("oracle", "duckdb"):
            answer_id_seq = Sequence(f"{__tablename__}_id_seq")
            id = Column(Integer, answer_id_seq, primary_key=True, autoincrement=True)
        else:
            id = Column(Integer, primary_key=True, autoincrement=True)
        question_id = Column(Integer, nullable=False)
        answer = Column(
            String(_get_table_len(table_len_config, "answer_answer")), nullable=False
        )
        answer_type = Column(Integer, nullable=False)

    class SessionTable(DynamicBase):
        """
        session table
        """

        __tablename__ = table_prefix + "_session"
        __table_args__ = {"extend_existing": True}

        if db_type in ("oracle", "duckdb"):
            session_id_seq = Sequence(f"{__tablename__}_id_seq", start=1)
            id = Column(
                Integer,
                session_id_seq,
                primary_key=True,
                autoincrement=True,
            )
        else:
            id = Column(Integer, primary_key=True, autoincrement=True)
        question_id = Column(Integer, nullable=False)
        session_id = Column(
            String(_get_table_len(table_len_config, "session_id")), nullable=False
        )
        session_question = Column(
            String(_get_table_len(table_len_config, "question_question")),
            nullable=False,
        )

    class QuestionDepTable(DynamicBase):
        """
        answer table
        """

        __tablename__ = table_prefix + "_question_dep"
        __table_args__ = {"extend_existing": True}

        if db_type in ("oracle", "duckdb"):
            question_dep_id_seq = Sequence(f"{__tablename__}_id_seq", start=1)
            id = Column(
                Integer, question_dep_id_seq, primary_key=True, autoincrement=True
            )
        else:
            id = Column(Integer, primary_key=True, autoincrement=True)
        question_id = Column(Integer, nullable=False)
        dep_name = Column(
            String(_get_table_len(table_len_config, "dep_name")), nullable=False
        )
        dep_data = Column(
            String(_get_table_len(table_len_config, "dep_data")), nullable=False
        )
        dep_type = Column(Integer, nullable=False)

    class ReportTable(DynamicBase):
        """
        report table
        """

        __tablename__ = table_prefix + "_report"
        __table_args__ = {"extend_existing": True}

        if db_type in ("oracle", "duckdb"):
            question_dep_id_seq = Sequence(f"{__tablename__}_id_seq", start=1)
            id = Column(
                Integer, question_dep_id_seq, primary_key=True, autoincrement=True
            )
        else:
            id = Column(Integer, primary_key=True, autoincrement=True)

        user_question = Column(
            String(_get_table_len(table_len_config, "question_question")),
            nullable=False,
        )
        cache_question_id = Column(
            Integer,
            nullable=False,
        )
        cache_question = Column(
            String(_get_table_len(table_len_config, "question_question")),
            nullable=False,
        )
        cache_answer = Column(
            String(_get_table_len(table_len_config, "answer_answer")), nullable=False
        )
        similarity = Column(Float, nullable=False)
        cache_delta_time = Column(Float, nullable=False)
        cache_time = Column(DateTime, default=datetime.now)
        extra = Column(
            String(_get_table_len(table_len_config, "question_question")),
            nullable=True,
        )

    return QuestionTable, AnswerTable, QuestionDepTable, SessionTable, ReportTable


class SQLStorage(CacheStorage):
    """
    Using sqlalchemy to manage SQLite, PostgreSQL, MySQL, MariaDB, SQL Server and Oracle.

    :param name: the name of the cache storage, it is support 'sqlite', 'postgresql', 'mysql', 'mariadb', 'sqlserver' and  'oracle' now.
    :type name: str
    :param sql_url: the url of the sql database for cache, such as '<db_type>+<db_driver>://<username>:<password>@<host>:<port>/<database>',
                    and the default value is related to the `cache_store` parameter,
                    'sqlite:///./sqlite.db' for 'sqlite',
                    'duckdb:///./duck.db' for 'duckdb',
                    'postgresql+psycopg2://postgres:123456@127.0.0.1:5432/postgres' for 'postgresql',
                    'mysql+pymysql://root:123456@127.0.0.1:3306/mysql' for 'mysql',
                    'mariadb+pymysql://root:123456@127.0.0.1:3307/mysql' for 'mariadb',
                    'mssql+pyodbc://sa:Strongpsw_123@127.0.0.1:1434/msdb?driver=ODBC+Driver+17+for+SQL+Server' for 'sqlserver',
                    'oracle+cx_oracle://oracle:123456@127.0.0.1:1521/?service_name=helowin&encoding=UTF-8&nencoding=UTF-8' for 'oracle'.
    :type sql_url: str
    :param table_name: the table name for sql database, defaults to 'gptcache'.
    :type table_name: str
    """

    def __init__(
        self,
        db_type: str = "sqlite",
        url: str = "sqlite:///./sqlite.db",
        table_name: str = "gptcache",
        table_len_config=None,
    ):
        if table_len_config is None:
            table_len_config = {}
        self._url = url
        self._ques, self._answer, self._ques_dep, self._session, self._report = get_models(
            table_name, db_type, table_len_config
        )
        self._engine = create_engine(self._url)
        self.Session = sessionmaker(bind=self._engine)  # pylint: disable=invalid-name
        self.create()

    def create(self):
        self._ques.__table__.create(bind=self._engine, checkfirst=True)
        self._answer.__table__.create(bind=self._engine, checkfirst=True)
        self._ques_dep.__table__.create(bind=self._engine, checkfirst=True)
        self._session.__table__.create(bind=self._engine, checkfirst=True)
        self._report.__table__.create(bind=self._engine, checkfirst=True)

    def _insert(self, data: CacheData, session: sqlalchemy.orm.Session) -> Column:
        ques_data = self._ques(
            question=data.question
            if isinstance(data.question, str)
            else data.question.content,
            embedding_data=data.embedding_data.tobytes()
            if data.embedding_data is not None
            else None,
        )
        session.add(ques_data)
        session.flush()
        if isinstance(data.question, Question) and data.question.deps is not None:
            all_deps = []
            for dep in data.question.deps:
                all_deps.append(
                    self._ques_dep(
                        question_id=ques_data.id,
                        dep_name=dep.name,
                        dep_data=dep.data,
                        dep_type=dep.dep_type,
                    )
                )
            session.add_all(all_deps)
        answers = data.answers if isinstance(data.answers, list) else [data.answers]
        all_data = []
        for answer in answers:
            answer_data = self._answer(
                question_id=ques_data.id,
                answer=answer.answer,
                answer_type=int(answer.answer_type),
            )
            all_data.append(answer_data)
        session.add_all(all_data)
        if data.session_id:
            session_data = self._session(
                question_id=ques_data.id,
                session_id=data.session_id,
                session_question=data.question
                if isinstance(data.question, str)
                else data.question.content,
            )
            session.add(session_data)
        return ques_data.id

    def batch_insert(self, all_data: List[CacheData]):
        ids = []
        with self.Session() as session:
            for data in all_data:
                ids.append(self._insert(data, session))
            session.commit()
        return ids

    def get_data_by_id(self, key: int) -> Optional[CacheData]:
        with self.Session() as session:
            qs = (
                session.query(self._ques)
                .filter(self._ques.id == key)
                .filter(self._ques.deleted == 0)
                .first()
            )
            if qs is None:
                return None
            last_access = qs.last_access
            qs.last_access = datetime.now()
            ans = (
                session.query(self._answer.answer, self._answer.answer_type)
                .filter(self._answer.question_id == qs.id)
                .all()
            )
            deps = (
                session.query(
                    self._ques_dep.dep_name,
                    self._ques_dep.dep_data,
                    self._ques_dep.dep_type,
                )
                .filter(self._ques_dep.question_id == qs.id)
                .all()
            )
            session_ids = (
                session.query(self._session.session_id)
                .filter(self._session.question_id == qs.id)
                .all()
            )
            res_ans = [(item.answer, item.answer_type) for item in ans]
            res_deps = [
                QuestionDep(item.dep_name, item.dep_data, item.dep_type)
                for item in deps
            ]
            session.commit()

            return CacheData(
                question=qs.question if not deps else Question(qs.question, res_deps),
                answers=res_ans,
                embedding_data=np.frombuffer(qs.embedding_data, dtype=np.float32),
                session_id=session_ids,
                create_on=qs.create_on,
                last_access=last_access,
            )

    def get_ids(self, deleted=True):
        state = -1 if deleted else 0
        with self.Session() as session:
            res = session.query(self._ques.id).filter(self._ques.deleted == state).all()
            return [item.id for item in res]

    def mark_deleted(self, keys):
        with self.Session() as session:
            session.query(self._ques).filter(self._ques.id.in_(keys)).update(
                {"deleted": -1}
            )
            session.commit()

    def clear_deleted_data(self):
        with self.Session() as session:
            objs = session.query(self._ques).filter(self._ques.deleted == -1)
            q_ids = [obj.id for obj in objs]
            session.query(self._answer).filter(
                self._answer.question_id.in_(q_ids)
            ).delete()
            session.query(self._ques_dep).filter(
                self._ques_dep.question_id.in_(q_ids)
            ).delete()
            session.query(self._session).filter(
                self._session.question_id.in_(q_ids)
            ).delete()
            objs.delete()
            session.commit()

    def count(self, state: int = 0, is_all: bool = False):
        with self.Session() as session:
            if is_all:
                return session.query(func.count(self._ques.id)).scalar()
            return (
                session.query(func.count(self._ques.id))
                .filter(self._ques.deleted == state)
                .scalar()
            )

    def add_session(self, question_id, session_id, session_question):
        with self.Session() as session:
            session_data = self._session(
                question_id=question_id,
                session_id=session_id,
                session_question=session_question,
            )
            session.add(session_data)
            session.commit()

    def delete_session(self, keys):
        with self.Session() as session:
            session.query(self._session).filter(self._session.id.in_(keys)).delete()
            session.commit()

    def list_sessions(self, session_id=None, key=None):
        with self.Session() as session:
            query = session.query(self._session)
            if session_id:
                query = query.filter(self._session.session_id == session_id)
            elif key:
                query = query.filter(self._session.question_id == key)
            return query.all()

    def report_cache(self, user_question, cache_question, cache_question_id, cache_answer, similarity_value, cache_delta_time):
        with self.Session() as session:
            report_data = self._report(
                user_question=user_question,
                cache_question=cache_question,
                cache_question_id=cache_question_id,
                cache_answer=cache_answer,
                similarity=similarity_value,
                cache_delta_time=cache_delta_time,
            )
            session.add(report_data)
            session.commit()

    def close(self):
        pass

    def count_answers(self):
        # for UT
        with self.Session() as session:
            return session.query(func.count(self._answer.id)).scalar()



import numpy as np

from gptcache.embedding.base import BaseEmbedding
from gptcache.utils import import_huggingface, import_torch

import_torch()
import_huggingface()

import torch  # pylint: disable=C0413
from transformers import AutoTokenizer, AutoModel  # pylint: disable=C0413


class Huggingface(BaseEmbedding):
    """Generate sentence embedding for given text using pretrained models from Huggingface transformers.

    :param model: model name, defaults to 'distilbert-base-uncased'.
    :type model: str

    Example:
        .. code-block:: python

            from gptcache.embedding import Huggingface

            test_sentence = 'Hello, world.'
            encoder = Huggingface(model='distilbert-base-uncased')
            embed = encoder.to_embeddings(test_sentence)

            test_sentence = '什么是Github'
            huggingface = Huggingface(model='uer/albert-base-chinese-cluecorpussmall')
            embed = huggingface.to_embeddings(test_sentence)
    """

    def __init__(self, model: str = "distilbert-base-uncased"):
        self.model = AutoModel.from_pretrained(model)
        self.model.eval()

        self.tokenizer = AutoTokenizer.from_pretrained(model)
        if not self.tokenizer.pad_token:
            self.tokenizer.pad_token = "[PAD]"
        try:
            self.__dimension = self.model.config.hidden_size
        except Exception:  # pylint: disable=W0703
            from transformers import AutoConfig  # pylint: disable=C0415

            config = AutoConfig.from_pretrained(model)
            self.__dimension = config.hidden_size

    def to_embeddings(self, data, **_):
        """Generate embedding given text input

        :param data: text in string.
        :type data: str

        :return: a text embedding in shape of (dim,).
        """
        if not isinstance(data, list):
            data = [data]
        inputs = self.tokenizer(
            data, padding=True, truncation=True, return_tensors="pt"
        )
        outs = self.model(**inputs).last_hidden_state
        emb = self.post_proc(outs, inputs).squeeze(0).detach().numpy()
        return np.array(emb).astype("float32")

    def post_proc(self, token_embeddings, inputs):
        attention_mask = inputs["attention_mask"]
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        sentence_embs = torch.sum(
            token_embeddings * input_mask_expanded, 1
        ) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)
        return sentence_embs

    @property
    def dimension(self):
        """Embedding dimension.

        :return: embedding dimension
        """
        return self.__dimension



import argparse
import json
import os
import zipfile
from typing import Optional

from gptcache import cache, Cache
from gptcache.adapter import openai
from gptcache.adapter.api import (
    get,
    put,
    init_similar_cache,
    init_similar_cache_from_config,
)
from gptcache.processor.pre import last_content
from gptcache.utils import import_fastapi, import_pydantic, import_starlette

import_fastapi()
import_pydantic()

from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import FileResponse
import uvicorn
from pydantic import BaseModel


app = FastAPI()
openai_cache: Optional[Cache] = None
cache_dir = ""
cache_file_key = ""


class CacheData(BaseModel):
    prompt: str
    answer: Optional[str] = ""


@app.get("/")
async def hello():
    return "hello gptcache server"


@app.post("/put")
async def put_cache(cache_data: CacheData) -> str:
    put(cache_data.prompt, cache_data.answer)
    return "successfully update the cache"


@app.post("/get")
async def get_cache(cache_data: CacheData) -> CacheData:
    result = get(cache_data.prompt)
    return CacheData(prompt=cache_data.prompt, answer=result)


@app.post("/flush")
async def flush_cache() -> str:
    cache.flush()
    return "successfully flush the cache"


@app.get("/cache_file")
async def get_cache_file(key: str = "") -> FileResponse:
    global cache_dir
    global cache_file_key
    if cache_dir == "":
        raise HTTPException(
            status_code=403,
            detail="the cache_dir was not specified when the service was initialized",
        )
    if cache_file_key == "":
        raise HTTPException(
            status_code=403,
            detail="the cache file can't be downloaded because the cache-file-key was not specified",
        )
    if cache_file_key != key:
        raise HTTPException(status_code=403, detail="the cache file key is wrong")
    zip_filename = cache_dir + ".zip"
    with zipfile.ZipFile(zip_filename, "w", compression=zipfile.ZIP_DEFLATED) as zipf:
        for root, _, files in os.walk(cache_dir):
            for file in files:
                zipf.write(os.path.join(root, file))
    return FileResponse(zip_filename)


@app.api_route(
    "/v1/chat/completions",
    methods=["POST", "OPTIONS"],
)
async def chat(request: Request):
    if openai_cache is None:
        raise HTTPException(
            status_code=500,
            detail=f"the gptcache server doesn't open the openai completes proxy",
        )

    import_starlette()
    from starlette.responses import StreamingResponse, JSONResponse

    openai_params = await request.json()
    is_stream = openai_params.get("stream", False)
    headers = request.headers
    auth_header = headers.get("authorization", None)
    openai_key = auth_header.split(" ")[1] if auth_header else ""
    cache_skip = openai_params.pop("cache_skip", False)
    if cache_skip is False:
        messages = openai_params.get("messages")
        if "/cache_skip " in messages[0]["content"]:
            cache_skip = True
            content0 = openai_params.get("messages")[0]["content"]
            openai_params.get("messages")[0]["content"] = str(content0).replace("/cache_skip ", "")
        elif "/cache_skip " in messages[-1]["content"]:
            cache_skip = True
            content0 = openai_params.get("messages")[-1]["content"]
            openai_params.get("messages")[-1]["content"] = str(content0).replace("/cache_skip ", "")
        print("cache_skip:", cache_skip)
    print("messages:", openai_params.get("messages"))
    try:
        if is_stream:
            def generate():
                for stream_response in openai.ChatCompletion.create(
                    cache_obj=openai_cache,
                    cache_skip=cache_skip,
                    api_key=openai_key,
                    **openai_params,
                ):
                    if stream_response == "[DONE]":
                        yield "data: [DONE]\n\n"
                        break
                    yield f"data: {json.dumps(stream_response)}\n\n"

            return StreamingResponse(generate(), media_type="text/event-stream")
        else:
            openai_response = openai.ChatCompletion.create(
                cache_obj=openai_cache,
                cache_skip=cache_skip,
                api_key=openai_key,
                **openai_params,
            )
            return JSONResponse(content=openai_response)
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"openai error: {e}")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-s", "--host", default="localhost", help="the hostname to listen on"
    )
    parser.add_argument(
        "-p", "--port", type=int, default=8000, help="the port to listen on"
    )
    parser.add_argument(
        "-d", "--cache-dir", default="gptcache_data", help="the cache data dir"
    )
    parser.add_argument("-k", "--cache-file-key", default="", help="the cache file key")
    parser.add_argument(
        "-f", "--cache-config-file", default=None, help="the cache config file"
    )
    parser.add_argument(
        "-o",
        "--openai",
        type=bool,
        default=False,
        help="whether to open the openai completes proxy",
    )
    parser.add_argument(
        "-of",
        "--openai-cache-config-file",
        default=None,
        help="the cache config file of the openai completes proxy",
    )

    args = parser.parse_args()
    global cache_dir
    global cache_file_key

    if args.cache_config_file:
        init_conf = init_similar_cache_from_config(config_dir=args.cache_config_file)
        cache_dir = init_conf.get("storage_config", {}).get("data_dir", "")
    else:
        init_similar_cache(args.cache_dir)
        cache_dir = args.cache_dir
    cache_file_key = args.cache_file_key

    if args.openai:
        global openai_cache
        openai_cache = Cache()
        if args.openai_cache_config_file:
            init_similar_cache_from_config(
                config_dir=args.openai_cache_config_file,
                cache_obj=openai_cache,
            )
        else:
            init_similar_cache(
                data_dir="openai_server_cache",
                pre_func=last_content,
                cache_obj=openai_cache,
            )

        import_starlette()
        from starlette.middleware.cors import CORSMiddleware

        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    uvicorn.run(app, host=args.host, port=args.port)


if __name__ == "__main__":
    main()



import re
import string
from typing import Dict, Any


def last_content(data: Dict[str, Any], **_: Dict[str, Any]) -> Any:
    """get the last content of the message list

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import last_content

            content = last_content({"messages": [{"content": "foo1"}, {"content": "foo2"}]})
            # content = "foo2"
    """
    return data.get("messages")[-1]["content"]


def last_content_without_prompt(data: Dict[str, Any], **params: Dict[str, Any]) -> Any:
    """get the last content of the message list without prompts content

    :param data: the user llm request data
    :type data: Dict[str, Any]
    :param params: the special gptcache params, like prompts param in the cache object
    :type params: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import last_content_without_prompt

            content = last_content_without_prompt(
                    {"messages": [{"content": "foo1"}, {"content": "foo2"}]}, prompts=["foo"]
                )
            # content = "2"
    """

    last_content_str = data.get("messages")[-1]["content"]
    prompts = params.get("prompts", [])
    if prompts is None:
        return last_content_str
    pattern = "|".join(prompts)
    new_content_str = re.sub(pattern, "", last_content_str)
    return new_content_str


def _get_pattern_value(pattern_str: str, value_str: str):
    literal_text_arr = []
    field_name_arr = []
    for literal_text, field_name, _, _ in string.Formatter().parse(pattern_str):
        literal_text_arr.append(literal_text)
        if field_name is not None:
            field_name_arr.append(
                field_name if field_name else str(len(field_name_arr))
            )

    pattern_values = {}
    last_end = 0
    for i, literal_text in enumerate(literal_text_arr):
        start = value_str.find(literal_text, last_end)
        if i == len(literal_text_arr) - 1:
            end = len(value_str)
        else:
            end = value_str.find(literal_text_arr[i + 1], start + 1)
        if start == -1 or end == -1:
            break
        start += len(literal_text)
        pattern_values[field_name_arr[i]] = value_str[start:end]
        last_end = end
    return pattern_values


def last_content_without_template(data: Dict[str, Any], **params: Dict[str, Any]) -> Any:
    """get the last content's template values of the message list without template content.

    When considering a cache agent or chain, the majority of the content consists of template content,
    while the essential information is simply a list of parameters within the template.
    In this way, the cache key is composed of a string made up of all the parameter values in the list.

    WARNING: Two parameters without intervals cannot appear in the template,
    for example: template = "{foo}{hoo}" is not supported,
    but template = "{foo}:{hoo}" is supported

    :param data: the user llm request data
    :type data: Dict[str, Any]

    :Example with str template:
        .. code-block:: python

            from gptcache import Config
            from gptcache.processor.pre import last_content_without_template

            template_obj = "tell me a joke about {subject}"
            prompt = template_obj.format(subject="animal")
            value = last_content_without_template(
                data={"messages": [{"content": prompt}]}, cache_config=Config(template=template_obj)
            )
            print(value)
            # ['animal']

    :Example with langchain template:
        .. code-block:: python

            from langchain import PromptTemplate

            from gptcache import Config
            from gptcache.processor.pre import last_content_without_template

            template_obj = PromptTemplate.from_template("tell me a joke about {subject}")
            prompt = template_obj.format(subject="animal")

            value = last_content_without_template(
                data={"messages": [{"content": prompt}]},
                cache_config=Config(template=template_obj.template),
            )
            print(value)
            # ['animal']

    NOTE: At present, only the simple PromptTemplate in langchain is supported.
    For ChatPromptTemplate, it needs to be adjusted according to the template array.
    If you need to use it, you need to pass in the final dialog template yourself.
    The reason why it cannot be advanced is that ChatPromptTemplate
    does not provide a method to directly return the template string.
    """
    last_content_str = data.get("messages")[-1]["content"]
    cache_config = params.get("cache_config", None)
    if not (cache_config and cache_config.template):
        return last_content_str

    pattern_value = _get_pattern_value(cache_config.template, last_content_str)
    return str(list(pattern_value.values()))


def all_content(data: Dict[str, Any], **_: Dict[str, Any]) -> Any:
    """get all content of the message list

    :param data: the user llm request data
    :type data: Dict[str, Any]

    :Example:
        .. code-block:: python

            from gptcache.processor.pre import all_content

            content = all_content(
                {"messages": [{"content": "foo1"}, {"content": "foo2"}]}
            )
            # content = "foo1\\nfoo2"
    """
    s = ""
    messages = data.get("messages")
    for i, message in enumerate(messages):
        if i == len(messages) - 1:
            s += message["content"]
        else:
            s += message["content"] + "\n"
    return s


def nop(data: Dict[str, Any], **_: Dict[str, Any]) -> Any:
    """do nothing of the llm request params

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import nop

            content = nop({"str": "hello"})
            # {"str": "hello"}
    """
    return data


def get_prompt(data: Dict[str, Any], **_: Dict[str, Any]) -> Any:
    """get the prompt of the llm request params

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_prompt

            content = get_prompt({"prompt": "foo"})
            # "foo"
    """
    return data.get("prompt")


def get_file_name(data: Dict[str, Any], **_: Dict[str, Any]) -> str:
    """get the file name of the llm request params

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_file_name

            file = open("test.txt", "a")
            content = get_file_name({"file": file})
            # "test.txt"
    """
    return data.get("file").name


def get_file_bytes(data: Dict[str, Any], **_: Dict[str, Any]) -> bytes:
    """get the file bytes of the llm request params

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_file_bytes

            content = get_file_bytes({"file": open("test.txt", "rb")})
    """
    return data.get("file").peek()


def get_input_str(data: Dict[str, Any], **_: Dict[str, Any]) -> str:
    """get the image and question str of the llm request params

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_input_str

            content = get_input_str({"input": {"image": open("test.png", "rb"), "question": "foo"}})
    """
    input_data = data.get("input")
    return str(input_data["image"].peek()) + input_data["question"]


def get_input_image_file_name(data: Dict[str, Any], **_: Dict[str, Any]) -> str:
    """get the image file name of the llm request params

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_input_image_file_name

            content = get_input_image_file_name({"input": {"image": open("test.png", "rb")}})
            # "test.png"
    """
    input_data = data.get("input")
    return input_data["image"].name


def get_image_question(data: Dict[str, Any], **_: Dict[str, Any]) -> str:  # pragma: no cover
    """get the image and question str of the llm request params

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_image_question

            content = get_image_question({"image": open("test.png", "rb"), "question": "foo"})
    """
    img = data.get("image")
    data_img = str(open(img, "rb").peek()) if isinstance(img, str) else str(img)  # pylint: disable=consider-using-with
    return data_img + data.get("question")


def get_image(data: Dict[str, Any], **_: Dict[str, Any]) -> str:  # pragma: no cover
    """get the image of the llm request params

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_image

            content = get_image({"image": open("test.png", "rb")})
            # "test.png"
    """
    return data.get("image")


def get_inputs(data: Dict[str, Any], **_: Dict[str, Any]):
    """get the inputs of the llm request params

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_inputs

            content = get_inputs({"inputs": "hello"})
            # "hello"
    """
    return data.get("inputs")


def get_messages_last_content(data: Dict[str, Any], **_: Any) -> str:
    """ get the last content of the llm request messages array

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_messages_last_content

            content = get_messages_last_content({"messages": [{"content": "hello"}, {"content": "world"}]})
            # "world"
    """
    return data.get("messages")[-1].content


def get_openai_moderation_input(data: Dict[str, Any], **_: Dict[str, Any]) -> str:
    """get the input param of the openai moderation request params

    :param data: the user openai moderation request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import get_openai_moderation_input

            content = get_openai_moderation_input({"input": ["hello", "world"]})
            # "['hello', 'world']"
    """

    return str(data.get("input"))


def concat_all_queries(data: Dict[str, Any], **params: Dict[str, Any]) -> Any:
    """

    :param data: the user llm request data
    :type data: Dict[str, Any]

    Example:
        .. code-block:: python

            from gptcache.processor.pre import concat_all_queries

            content = concat_all_queries({"messages": [{"role": "system", "content": "hello"},
                {"role": "user", "content": "world"},
                {"role": "assistant", "content": "alice"}]})

    """
    cache_config = params.get("cache_config", None)
    skip_list = cache_config.skip_list
    context_len = cache_config.context_len
    context_len = context_len * 2
    s = ""
    messages = data.get("messages")
    length = min(context_len, len(messages))
    messages = messages[len(messages) - length:]
    for i, message in enumerate(messages):
        if message["role"] in skip_list:
            continue
        if i == len(messages) - 1:
            s += f'{message["role"].upper()}: {message["content"]}'
        else:
            s += f'{message["role"].upper()}: {message["content"]}\n'
    return s



from typing import List

import numpy as np

from gptcache.manager.vector_data.base import VectorBase, VectorData
from gptcache.utils import import_redis
from gptcache.utils.log import gptcache_log

import_redis()

# pylint: disable=C0413
from redis.commands.search.indexDefinition import IndexDefinition, IndexType
from redis.commands.search.query import Query
from redis.commands.search.field import TagField, VectorField
from redis.client import Redis


class RedisVectorStore(VectorBase):
    """ vector store: Redis

    :param host: redis host, defaults to "localhost".
    :type host: str
    :param port: redis port, defaults to "6379".
    :type port: str
    :param username: redis username, defaults to "".
    :type username: str
    :param password: redis password, defaults to "".
    :type password: str
    :param dimension: the dimension of the vector, defaults to 0.
    :type dimension: int
    :param collection_name: the name of the index for Redis, defaults to "gptcache".
    :type collection_name: str
    :param top_k: the number of the vectors results to return, defaults to 1.
    :type top_k: int

    Example:
        .. code-block:: python

            from gptcache.manager import VectorBase

            vector_base = VectorBase("redis", dimension=10)
    """
    def __init__(
        self,
        host: str = "localhost",
        port: str = "6379",
        username: str = "",
        password: str = "",
        dimension: int = 0,
        collection_name: str = "gptcache",
        top_k: int = 1,
        namespace: str = "",
    ):
        self._client = Redis(
            host=host, port=int(port), username=username, password=password
        )
        self.top_k = top_k
        self.dimension = dimension
        self.collection_name = collection_name
        self.namespace = namespace
        self.doc_prefix = f"{self.namespace}doc:"  # Prefix with the specified namespace
        self._create_collection(collection_name)

    def _check_index_exists(self, index_name: str) -> bool:
        """Check if Redis index exists."""
        try:
            self._client.ft(index_name).info()
        except:  # pylint: disable=W0702
            gptcache_log.info("Index does not exist")
            return False
        gptcache_log.info("Index already exists")
        return True

    def _create_collection(self, collection_name):
        if self._check_index_exists(collection_name):
            gptcache_log.info(
                "The %s already exists, and it will be used directly", collection_name
            )
        else:
            schema = (
                TagField("tag"),  # Tag Field Name
                VectorField(
                    "vector",  # Vector Field Name
                    "FLAT",
                    {  # Vector Index Type: FLAT or HNSW
                        "TYPE": "FLOAT32",  # FLOAT32 or FLOAT64
                        "DIM": self.dimension,  # Number of Vector Dimensions
                        "DISTANCE_METRIC": "COSINE",  # Vector Search Distance Metric
                    },
                ),
            )
            definition = IndexDefinition(
                prefix=[self.doc_prefix], index_type=IndexType.HASH
            )

            # create Index
            self._client.ft(collection_name).create_index(
                fields=schema, definition=definition
            )

    def mul_add(self, datas: List[VectorData]):
        pipe = self._client.pipeline()

        for data in datas:
            key: int = data.id
            obj = {
                "vector": data.data.astype(np.float32).tobytes(),
            }
            pipe.hset(f"{self.doc_prefix}{key}", mapping=obj)

        pipe.execute()

    def search(self, data: np.ndarray, top_k: int = -1):
        query = (
            Query(
                f"*=>[KNN {top_k if top_k > 0 else self.top_k} @vector $vec as score]"
            )
            .sort_by("score")
            .return_fields("id", "score")
            .paging(0, top_k if top_k > 0 else self.top_k)
            .dialect(2)
        )
        query_params = {"vec": data.astype(np.float32).tobytes()}
        results = (
            self._client.ft(self.collection_name)
            .search(query, query_params=query_params)
            .docs
        )
        return [(float(result.score), int(result.id[len(self.doc_prefix):])) for result in results]

    def rebuild(self, ids=None) -> bool:
        pass

    def delete(self, ids) -> None:
        pipe = self._client.pipeline()
        for data_id in ids:
            pipe.delete(f"{self.doc_prefix}{data_id}")
        pipe.execute()



import numpy as np

from gptcache.embedding.base import BaseEmbedding
from gptcache.utils import import_langchain

import_langchain()

from langchain.embeddings.base import Embeddings  # pylint: disable=C0413


class LangChain(BaseEmbedding):
    """Generate text embedding for given text using LangChain

    :param embeddings: the LangChain Embeddings object.
    :type embeddings: Embeddings
    :param dimension: The vector dimension after embedding is calculated by calling embed once by default.
     If you confirm the dimension, you can assign a value to this parameter to reduce this request.
    :type dimension: int

    Example:
        .. code-block:: python

            from gptcache.embedding import LangChain
            from langchain.embeddings.openai import OpenAIEmbeddings

            test_sentence = 'Hello, world.'
            embeddings = OpenAIEmbeddings(model="your-embeddings-deployment-name")
            encoder = LangChain(embeddings=embeddings)
            embed = encoder.to_embeddings(test_sentence)
    """

    def __init__(self, embeddings: Embeddings, dimension: int = 0):
        self._embeddings: Embeddings = embeddings
        self._dimension: int = (
            dimension if dimension != 0 else len(self._embeddings.embed_query("foo"))
        )

    def to_embeddings(self, data, **kwargs):
        vector_data = self._embeddings.embed_query(data)
        return np.array(vector_data).astype("float32")

    @property
    def dimension(self) -> int:
        return self._dimension



import time

import numpy as np

from gptcache import cache
from gptcache.processor.post import temperature_softmax
from gptcache.utils.error import NotInitError
from gptcache.utils.log import gptcache_log
from gptcache.utils.time import time_cal


def adapt(llm_handler, cache_data_convert, update_cache_callback, *args, **kwargs):
    """Adapt to different llm

    :param llm_handler: LLM calling method, when the cache misses, this function will be called
    :param cache_data_convert: When the cache hits, convert the answer in the cache to the format of the result returned by llm
    :param update_cache_callback: If the cache misses, after getting the result returned by llm, save the result to the cache
    :param args: llm args
    :param kwargs: llm kwargs
    :return: llm result
    """
    start_time = time.time()
    search_only_flag = kwargs.pop("search_only", False)
    user_temperature = "temperature" in kwargs
    user_top_k = "top_k" in kwargs
    temperature = kwargs.pop("temperature", 0.0)
    chat_cache = kwargs.pop("cache_obj", cache)
    session = kwargs.pop("session", None)
    require_object_store = kwargs.pop("require_object_store", False)
    if require_object_store:
        assert chat_cache.data_manager.o, "Object store is required for adapter."
    if not chat_cache.has_init:
        raise NotInitError()
    cache_enable = chat_cache.cache_enable_func(*args, **kwargs)
    context = kwargs.pop("cache_context", {})
    embedding_data = None
    # you want to retry to send the request to chatgpt when the cache is negative

    if 0 < temperature < 2:
        cache_skip_options = [True, False]
        prob_cache_skip = [0, 1]
        cache_skip = kwargs.pop(
            "cache_skip",
            temperature_softmax(
                messages=cache_skip_options,
                scores=prob_cache_skip,
                temperature=temperature,
            ),
        )
    elif temperature >= 2:
        cache_skip = kwargs.pop("cache_skip", True)
    else:  # temperature <= 0
        cache_skip = kwargs.pop("cache_skip", False)
    cache_factor = kwargs.pop("cache_factor", 1.0)
    pre_embedding_res = time_cal(
        chat_cache.pre_embedding_func,
        func_name="pre_process",
        report_func=chat_cache.report.pre,
    )(
        kwargs,
        extra_param=context.get("pre_embedding_func", None),
        prompts=chat_cache.config.prompts,
        cache_config=chat_cache.config,
    )
    if isinstance(pre_embedding_res, tuple):
        pre_store_data = pre_embedding_res[0]
        pre_embedding_data = pre_embedding_res[1]
    else:
        pre_store_data = pre_embedding_res
        pre_embedding_data = pre_embedding_res

    if chat_cache.config.input_summary_len is not None:
        pre_embedding_data = _summarize_input(
            pre_embedding_data, chat_cache.config.input_summary_len
        )

    if cache_enable:
        embedding_data = time_cal(
            chat_cache.embedding_func,
            func_name="embedding",
            report_func=chat_cache.report.embedding,
        )(pre_embedding_data, extra_param=context.get("embedding_func", None))
    if cache_enable and not cache_skip:
        search_data_list = time_cal(
            chat_cache.data_manager.search,
            func_name="search",
            report_func=chat_cache.report.search,
        )(
            embedding_data,
            extra_param=context.get("search_func", None),
            top_k=kwargs.pop("top_k", 5)
            if (user_temperature and not user_top_k)
            else kwargs.pop("top_k", -1),
        )
        if search_data_list is None:
            search_data_list = []
        cache_answers = []
        similarity_threshold = chat_cache.config.similarity_threshold
        min_rank, max_rank = chat_cache.similarity_evaluation.range()
        rank_threshold = (max_rank - min_rank) * similarity_threshold * cache_factor
        rank_threshold = (
            max_rank
            if rank_threshold > max_rank
            else min_rank
            if rank_threshold < min_rank
            else rank_threshold
        )
        for search_data in search_data_list:
            cache_data = time_cal(
                chat_cache.data_manager.get_scalar_data,
                func_name="get_data",
                report_func=chat_cache.report.data,
            )(
                search_data,
                extra_param=context.get("get_scalar_data", None),
                session=session,
            )
            if cache_data is None:
                continue

            # cache consistency check
            if chat_cache.config.data_check:
                is_healthy = cache_health_check(
                    chat_cache.data_manager.v,
                    {
                        "embedding": cache_data.embedding_data,
                        "search_result": search_data,
                    },
                )
                if not is_healthy:
                    continue

            if "deps" in context and hasattr(cache_data.question, "deps"):
                eval_query_data = {
                    "question": context["deps"][0]["data"],
                    "embedding": None,
                }
                eval_cache_data = {
                    "question": cache_data.question.deps[0].data,
                    "answer": cache_data.answers[0].answer,
                    "search_result": search_data,
                    "cache_data": cache_data,
                    "embedding": None,
                }
            else:
                eval_query_data = {
                    "question": pre_store_data,
                    "embedding": embedding_data,
                }

                eval_cache_data = {
                    "question": cache_data.question,
                    "answer": cache_data.answers[0].answer,
                    "search_result": search_data,
                    "cache_data": cache_data,
                    "embedding": cache_data.embedding_data,
                }
            rank = time_cal(
                chat_cache.similarity_evaluation.evaluation,
                func_name="evaluation",
                report_func=chat_cache.report.evaluation,
            )(
                eval_query_data,
                eval_cache_data,
                extra_param=context.get("evaluation_func", None),
            )
            gptcache_log.debug(
                "similarity: [user question] %s, [cache question] %s, [value] %f",
                pre_store_data,
                cache_data.question,
                rank,
            )
            if rank_threshold <= rank:
                cache_answers.append(
                    (float(rank), cache_data.answers[0].answer, search_data, cache_data)
                )
                chat_cache.data_manager.hit_cache_callback(search_data)
        cache_answers = sorted(cache_answers, key=lambda x: x[0], reverse=True)
        answers_dict = dict((d[1], d) for d in cache_answers)
        if len(cache_answers) != 0:
            hit_callback = kwargs.pop("hit_callback", None)
            if hit_callback and callable(hit_callback):
                factor = max_rank - min_rank
                hit_callback([(d[3].question, d[0] / factor if factor else d[0]) for d in cache_answers])
            def post_process():
                if chat_cache.post_process_messages_func is temperature_softmax:
                    return_message = chat_cache.post_process_messages_func(
                        messages=[t[1] for t in cache_answers],
                        scores=[t[0] for t in cache_answers],
                        temperature=temperature,
                    )
                else:
                    return_message = chat_cache.post_process_messages_func(
                        [t[1] for t in cache_answers]
                    )
                return return_message

            return_message = time_cal(
                post_process,
                func_name="post_process",
                report_func=chat_cache.report.post,
            )()
            chat_cache.report.hint_cache()
            cache_whole_data = answers_dict.get(str(return_message))
            if session and cache_whole_data:
                chat_cache.data_manager.add_session(
                    cache_whole_data[2], session.name, pre_embedding_data
                )
            if cache_whole_data and not chat_cache.config.disable_report:
                # user_question / cache_question / cache_question_id / cache_answer / similarity / consume time/ time
                report_cache_data = cache_whole_data[3]
                report_search_data = cache_whole_data[2]
                chat_cache.data_manager.report_cache(
                    pre_store_data if isinstance(pre_store_data, str) else "",
                    report_cache_data.question
                    if isinstance(report_cache_data.question, str)
                    else "",
                    report_search_data[1],
                    report_cache_data.answers[0].answer
                    if isinstance(report_cache_data.answers[0].answer, str)
                    else "",
                    cache_whole_data[0],
                    round(time.time() - start_time, 6),
                )
            return cache_data_convert(return_message)

    next_cache = chat_cache.next_cache
    if next_cache:
        kwargs["cache_obj"] = next_cache
        kwargs["cache_context"] = context
        kwargs["cache_skip"] = cache_skip
        kwargs["cache_factor"] = cache_factor
        kwargs["search_only"] = search_only_flag
        llm_data = adapt(
            llm_handler, cache_data_convert, update_cache_callback, *args, **kwargs
        )
    else:
        if search_only_flag:
            # cache miss
            return None
        llm_data = time_cal(
            llm_handler, func_name="llm_request", report_func=chat_cache.report.llm
        )(*args, **kwargs)

    if not llm_data:
        return None

    if cache_enable:
        try:

            def update_cache_func(handled_llm_data, question=None):
                if question is None:
                    question = pre_store_data
                else:
                    question.content = pre_store_data
                time_cal(
                    chat_cache.data_manager.save,
                    func_name="save",
                    report_func=chat_cache.report.save,
                )(
                    question,
                    handled_llm_data,
                    embedding_data,
                    extra_param=context.get("save_func", None),
                    session=session,
                )
                if (
                    chat_cache.report.op_save.count > 0
                    and chat_cache.report.op_save.count % chat_cache.config.auto_flush
                    == 0
                ):
                    chat_cache.flush()

            llm_data = update_cache_callback(
                llm_data, update_cache_func, *args, **kwargs
            )
        except Exception as e:  # pylint: disable=W0703
            gptcache_log.warning("failed to save the data to cache, error: %s", e)
    return llm_data


async def aadapt(
    llm_handler, cache_data_convert, update_cache_callback, *args, **kwargs
):
    """Simple copy of the 'adapt' method to different llm for 'async llm function'

    :param llm_handler: Async LLM calling method, when the cache misses, this function will be called
    :param cache_data_convert: When the cache hits, convert the answer in the cache to the format of the result returned by llm
    :param update_cache_callback: If the cache misses, after getting the result returned by llm, save the result to the cache
    :param args: llm args
    :param kwargs: llm kwargs
    :return: llm result
    """
    start_time = time.time()
    user_temperature = "temperature" in kwargs
    user_top_k = "top_k" in kwargs
    temperature = kwargs.pop("temperature", 0.0)
    chat_cache = kwargs.pop("cache_obj", cache)
    session = kwargs.pop("session", None)
    require_object_store = kwargs.pop("require_object_store", False)
    if require_object_store:
        assert chat_cache.data_manager.o, "Object store is required for adapter."
    if not chat_cache.has_init:
        raise NotInitError()
    cache_enable = chat_cache.cache_enable_func(*args, **kwargs)
    context = kwargs.pop("cache_context", {})
    embedding_data = None
    # you want to retry to send the request to chatgpt when the cache is negative

    if 0 < temperature < 2:
        cache_skip_options = [True, False]
        prob_cache_skip = [0, 1]
        cache_skip = kwargs.pop(
            "cache_skip",
            temperature_softmax(
                messages=cache_skip_options,
                scores=prob_cache_skip,
                temperature=temperature,
            ),
        )
    elif temperature >= 2:
        cache_skip = kwargs.pop("cache_skip", True)
    else:  # temperature <= 0
        cache_skip = kwargs.pop("cache_skip", False)
    cache_factor = kwargs.pop("cache_factor", 1.0)
    pre_embedding_res = time_cal(
        chat_cache.pre_embedding_func,
        func_name="pre_process",
        report_func=chat_cache.report.pre,
    )(
        kwargs,
        extra_param=context.get("pre_embedding_func", None),
        prompts=chat_cache.config.prompts,
        cache_config=chat_cache.config,
    )
    if isinstance(pre_embedding_res, tuple):
        pre_store_data = pre_embedding_res[0]
        pre_embedding_data = pre_embedding_res[1]
    else:
        pre_store_data = pre_embedding_res
        pre_embedding_data = pre_embedding_res

    if chat_cache.config.input_summary_len is not None:
        pre_embedding_data = _summarize_input(
            pre_embedding_data, chat_cache.config.input_summary_len
        )

    if cache_enable:
        embedding_data = time_cal(
            chat_cache.embedding_func,
            func_name="embedding",
            report_func=chat_cache.report.embedding,
        )(pre_embedding_data, extra_param=context.get("embedding_func", None))
    if cache_enable and not cache_skip:
        search_data_list = time_cal(
            chat_cache.data_manager.search,
            func_name="search",
            report_func=chat_cache.report.search,
        )(
            embedding_data,
            extra_param=context.get("search_func", None),
            top_k=kwargs.pop("top_k", 5)
            if (user_temperature and not user_top_k)
            else kwargs.pop("top_k", -1),
        )
        if search_data_list is None:
            search_data_list = []
        cache_answers = []
        similarity_threshold = chat_cache.config.similarity_threshold
        min_rank, max_rank = chat_cache.similarity_evaluation.range()
        rank_threshold = (max_rank - min_rank) * similarity_threshold * cache_factor
        rank_threshold = (
            max_rank
            if rank_threshold > max_rank
            else min_rank
            if rank_threshold < min_rank
            else rank_threshold
        )
        for search_data in search_data_list:
            cache_data = time_cal(
                chat_cache.data_manager.get_scalar_data,
                func_name="get_data",
                report_func=chat_cache.report.data,
            )(
                search_data,
                extra_param=context.get("get_scalar_data", None),
                session=session,
            )
            if cache_data is None:
                continue

            if "deps" in context and hasattr(cache_data.question, "deps"):
                eval_query_data = {
                    "question": context["deps"][0]["data"],
                    "embedding": None,
                }
                eval_cache_data = {
                    "question": cache_data.question.deps[0].data,
                    "answer": cache_data.answers[0].answer,
                    "search_result": search_data,
                    "cache_data": cache_data,
                    "embedding": None,
                }
            else:
                eval_query_data = {
                    "question": pre_store_data,
                    "embedding": embedding_data,
                }

                eval_cache_data = {
                    "question": cache_data.question,
                    "answer": cache_data.answers[0].answer,
                    "search_result": search_data,
                    "cache_data": cache_data,
                    "embedding": cache_data.embedding_data,
                }
            rank = time_cal(
                chat_cache.similarity_evaluation.evaluation,
                func_name="evaluation",
                report_func=chat_cache.report.evaluation,
            )(
                eval_query_data,
                eval_cache_data,
                extra_param=context.get("evaluation_func", None),
            )
            gptcache_log.debug(
                "similarity: [user question] %s, [cache question] %s, [value] %f",
                pre_store_data,
                cache_data.question,
                rank,
            )
            if rank_threshold <= rank:
                cache_answers.append(
                    (float(rank), cache_data.answers[0].answer, search_data, cache_data)
                )
                chat_cache.data_manager.hit_cache_callback(search_data)
        cache_answers = sorted(cache_answers, key=lambda x: x[0], reverse=True)
        answers_dict = dict((d[1], d) for d in cache_answers)
        if len(cache_answers) != 0:
            def post_process():
                if chat_cache.post_process_messages_func is temperature_softmax:
                    return_message = chat_cache.post_process_messages_func(
                        messages=[t[1] for t in cache_answers],
                        scores=[t[0] for t in cache_answers],
                        temperature=temperature,
                    )
                else:
                    return_message = chat_cache.post_process_messages_func(
                        [t[1] for t in cache_answers]
                    )
                return return_message

            return_message = time_cal(
                post_process,
                func_name="post_process",
                report_func=chat_cache.report.post,
            )()
            chat_cache.report.hint_cache()
            cache_whole_data = answers_dict.get(str(return_message))
            if session and cache_whole_data:
                chat_cache.data_manager.add_session(
                    cache_whole_data[2], session.name, pre_embedding_data
                )
            if cache_whole_data:
                # user_question / cache_question / cache_question_id / cache_answer / similarity / consume time/ time
                report_cache_data = cache_whole_data[3]
                report_search_data = cache_whole_data[2]
                chat_cache.data_manager.report_cache(
                    pre_store_data if isinstance(pre_store_data, str) else "",
                    report_cache_data.question
                    if isinstance(report_cache_data.question, str)
                    else "",
                    report_search_data[1],
                    report_cache_data.answers[0].answer
                    if isinstance(report_cache_data.answers[0].answer, str)
                    else "",
                    cache_whole_data[0],
                    round(time.time() - start_time, 6),
                )
            return cache_data_convert(return_message)

    next_cache = chat_cache.next_cache
    if next_cache:
        kwargs["cache_obj"] = next_cache
        kwargs["cache_context"] = context
        kwargs["cache_skip"] = cache_skip
        kwargs["cache_factor"] = cache_factor
        llm_data = adapt(
            llm_handler, cache_data_convert, update_cache_callback, *args, **kwargs
        )
    else:
        llm_data = await llm_handler(*args, **kwargs)

    if cache_enable:
        try:

            def update_cache_func(handled_llm_data, question=None):
                if question is None:
                    question = pre_store_data
                else:
                    question.content = pre_store_data
                time_cal(
                    chat_cache.data_manager.save,
                    func_name="save",
                    report_func=chat_cache.report.save,
                )(
                    question,
                    handled_llm_data,
                    embedding_data,
                    extra_param=context.get("save_func", None),
                    session=session,
                )
                if (
                    chat_cache.report.op_save.count > 0
                    and chat_cache.report.op_save.count % chat_cache.config.auto_flush
                    == 0
                ):
                    chat_cache.flush()
            llm_data = update_cache_callback(
                llm_data, update_cache_func, *args, **kwargs
            )
        except Exception:  # pylint: disable=W0703
            gptcache_log.error("failed to save the data to cache", exc_info=True)
    return llm_data


_input_summarizer = None


def _summarize_input(text, text_length):
    if len(text) <= text_length:
        return text

    # pylint: disable=import-outside-toplevel
    from gptcache.processor.context.summarization_context import (
        SummarizationContextProcess,
    )

    global _input_summarizer
    if _input_summarizer is None:
        _input_summarizer = SummarizationContextProcess()
    summarization = _input_summarizer.summarize_to_sentence([text], text_length)
    return summarization


def cache_health_check(vectordb, cache_dict):
    """This function checks if the embedding
    from vector store matches one in cache store.
    If cache store and vector store are out of
    sync with each other, cache retrieval can
    be incorrect.
    If this happens, force the similary score
    to the lowerest possible value.
    """
    emb_in_cache = cache_dict["embedding"]
    _, data_id = cache_dict["search_result"]
    emb_in_vec = vectordb.get_embeddings(data_id)
    flag = np.all(emb_in_cache == emb_in_vec)
    if not flag:
        gptcache_log.critical("Cache Store and Vector Store are out of sync!!!")
        # 0: identical, inf: different
        cache_dict["search_result"] = (
            np.inf,
            data_id,
        )
        # self-healing by replacing entry
        # in the vec store with the one
        # from cache store by the same
        # entry_id.
        vectordb.update_embeddings(
            data_id,
            emb=cache_dict["embedding"],
        )
    return flag
